/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#define	PARAM_MAX_TTL			0
#define	PARAM_DEFAULT_TTL		1
#define	PARAM_ROUTING_TABLE_SIZE	2
#define	PARAM_MAX_CONNECTIONS		3
#define	PARAM_MAX_RESULTS		4
#define	PARAM_MAX_HOPS			5
#define	PARAM_MAX_PUSH_REQUESTS		6
#define	PARAM_MAX_UPLOADS		7
#define	PARAM_MAX_DOWNLOADS		8
#define	PARAM_MAX_HOSTS			9
#define	PARAM_PORT			10
#define	PARAM_SPEED			11
#define	PARAM_MAX_MONITOR_LENGTH	12
#define	PARAM_AUTO_CLEAR_COMPLETE	13
#define	PARAM_AUTO_CONNECTIONS		14
#define	PARAM_AUTO_PUSH			15
#define	PARAM_AUTO_CLEAR_RESULTS	16
#define	PARAM_MAX_HOST_CON_FAIL		17
#define	PARAM_AUTO_SAVE_HOSTS		18
#define	PARAM_DAEMON_PORT		19
#define PARAM_CLIENTID			20
#define PARAM_DOWNLOAD_DIR		21
#define PARAM_MAX_FILES_SHARED		22
#define PARAM_MAX_CON_ERRORS		23
#define PARAM_IGNORE_LOCAL_IP		24
#define PARAM_NOTIFY_RESULTS		25
#define PARAM_CACHE_ENTRIES		26
#define PARAM_PACKET_RATIO		27
#define PARAM_INVALID			28
/*
#define PARAM_
#define PARAM_
*/

class Backend
{
	private	:
		FServer			*fileServer;
		GnutellaRouter		*GRouter;
		ControlCentre		*MCC;
		Networker		*netMan;
		HostCatcher		*myHostCatcher;
		DB			*connections;
		DB			*hosts;
		Mesg			*ourInit;
#ifndef ROUTER
		SearchManager		*searchMan;
		Downloader		*leechMan;
		Monitor			*bigBrother;
		Mesg			*sign;
		unsigned int		hostsOnNet;
		unsigned long int	filesOnNet;
		unsigned long int	kbsOnNet;
		int			handlePendingOutgoingConnections();
#endif
		int			hostIndex;
		char			ourClientID[CLIENTID_LENGTH];
		void			handleInitResp(GnutellaPacket *p);
		void			handleSearch(GnutellaPacket *p);
		int			handleNewPacket(GnutellaPacket *p);
		void			handleNewConnection(Connection *c);
		int			handleInit(GnutellaPacket *p);
		int			handlePush(GnutellaPacket *p);
		int			autoOpenConnections();
		int			autoPush();
		
	public :
		Backend(ControlCentre *newMCC);

		~Backend();


		/*
		 * GnutellaNet interface
		 */

		int 		storeHost(char *dotted, int tries, char *cid);

		int 		deleteHost(int i);

		int 		connectToHost(int i);

		int 		closeConnection(int i);

		int 		closeHost(int i);

		GnutellaConnection 	*getConnection(int i);

		Host 		*getHost(int i);

		int 		pollGnutellaNet();

		int 		sendInit();	
		
		int 		startListening();	
		
		int 		writeHosts(char *fileName);

		int 		writeShares(char *fileName);

		int 		reload();

#ifndef ROUTER
		int		writeOptions(char *fileName);
#endif

		/*
		 * Einstein interface
		 */

		 void		set_AUTO_CONNECTIONS(int nc);

		 void		set_AUTO_PUSH(int yn);

		 void		set_AUTO_CLEAR_COMPLETE(int yn);

#ifndef ROUTER
		/*
		 * Monitor interface
		 */

		 Query		*getQueries();

		unsigned int	getQueryCount();
#endif

		/*
		 * FileServer interface
		 */
		int 		addPath(char *newPath);

		int 		delPath(char *delPath);

		int 		delAllPaths();

		int 		scanDirs();

		int 		killUpload(int idx);

		Path		*getPathList();

		TransferStatus 	*getUploads();

		int		getFilesShared();

		int		getSizeShared();

		unsigned int	getLocalSearches();

		unsigned int	getCacheHits();

		unsigned int	getCacheEntries();

		unsigned int	getCacheSize();

		unsigned int	getSearchArraySize();

		unsigned int	getHashArraySize();

#ifndef ROUTER

		/*
		 * SearchManager interface
		 */

		int 		launchQuery(int speed, char *findMe);

		SearchResult 	*getOneResult(int i);

		void		clearResults();

		unsigned int	getSMQueries();

		unsigned int	getSMResultCount();

		/*
		 * Download Manager Interface
		 */
		int 		startDownload(int idx);

		int 		continueDownload(int idx);

		int 		killDownload(int idx);

		int		pushDownload(int idx);

		TransferStatus *getDownloads();

		unsigned int	getDLBytesDownloaded();
#endif

		/*
		 * Controls
		 */

                int 		get_INT_PARAM(int which);

                void		get_STR_PARAM(int which, char *buf);

                void		set_INT_PARAM(int which, int val);

                void		set_STR_PARAM(int which, char *buf);

		/*
		 * Special Functions
		 */
		
		void		newClientID(char *newCID);

		/*
		 * General Stats Stats
		 */

#ifndef ROUTER
		unsigned int		getHostsOnNet();

		unsigned long int	getFilesOnNet();

		unsigned long int	getSizeOnNet();
#endif

		/*
		 * GnutellaRouter Stats
		 */
		unsigned int		getGRInitPackets();

		unsigned int		getGRInitRespPackets();

		unsigned int		getGRPushPackets();

		unsigned int		getGRSearchPackets();

		unsigned int		getGRSearchRespPackets();

		unsigned int		getGRDoublePackets();

		unsigned int		getGRRespMatches();

		unsigned int		getGRRespMisses();

		/*
		 * HostCatcher Stats
		 */
		unsigned int		getHCHostsCaught();

		/*
		 * Networker Stats
		 */
		unsigned int		getNWBytesIn();

		unsigned int		getNWBytesOut();
};
