/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
*/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "Socket.h"
#include "Buffer.h"

Buffer::Buffer()
{
	writePtr	= 0;
	readPtr		= 0;
	full		= 0;
	size		= BUFFER_STD_SIZE;
	buffer		= (char *)malloc(size);
}

Buffer::Buffer(unsigned int sz)
{
	writePtr	= 0;
	readPtr		= 0;
	full		= 0;
	if (sz > 0)
		size		= sz;
	else
		size		= BUFFER_STD_SIZE;
	buffer		= (char *)malloc(size);
}

Buffer::~Buffer()
{
	free(buffer);
}

Buffer *Buffer::copyBuffer()
{
	Buffer	*b;

	b	= new Buffer(size);

	if (!this->bytesInBuffer())
		return b;
	if (writePtr > readPtr)
		b->add(buffer + readPtr, this->bytesInBuffer());
	else
	{
		b->add(buffer + readPtr, size - readPtr);
		b->add(buffer, writePtr);
	}

	return b;
}

int Buffer::add(char *src, unsigned int len)
{
	int	bytesAdded;

	if (full)
		return 0;

	if (readPtr > writePtr)
	{
		if (len < readPtr - writePtr)
			bytesAdded	= len;
		else
			bytesAdded	= readPtr - writePtr;

		memcpy(buffer + writePtr, src, bytesAdded);
		writePtr	+= bytesAdded;
	}
	else if (len < size - writePtr)
	{
		bytesAdded	= len;
		memcpy(buffer + writePtr, src, bytesAdded);
		writePtr	+= bytesAdded;
	}
	else
	{
		int	ba;

		ba	= bytesAdded	= size - writePtr;
		memcpy(buffer + writePtr, src, ba);
		if ((len - ba) < readPtr)
		{
			ba	= len - ba;
			memcpy(buffer, src + bytesAdded, ba);
		}
		else
		{
			ba	= readPtr;
			memcpy(buffer, src + bytesAdded, ba);
		}
		bytesAdded	+= ba;
		writePtr	= ba;
	}

	if (writePtr >= size)
		writePtr	= 0;

	if (	(bytesAdded	> 0)	&&
		(writePtr == readPtr)	)
		full	= 1;

	return bytesAdded;
}

int Buffer::skip(unsigned int len)
{
	if ( (readPtr == writePtr) && (!full) )
		return 0;

	if (writePtr > readPtr)
	{
		if (len > writePtr - readPtr)
			len	= writePtr - readPtr;
		readPtr	+= len;
	}
	else if (len <= size - readPtr)
		readPtr	+= len;
	else
	{
		int	skipped;

		skipped	= size - readPtr;
		readPtr	= 0;
		if (len - skipped > writePtr)
		{
			readPtr	+= writePtr;
			len 	= skipped + writePtr;
		}
		else
			readPtr	+= len - skipped;
	}

	if (	(full)	&&
		(len > 0)	)
		full	= 0;

	return len;
}

int Buffer::rewind(unsigned int len)
{
	if (	((writePtr == readPtr) && (!full))	||
		(!len)	)
		return 0;

	if (writePtr > readPtr)
	{
		if (len <= readPtr)
		{
			readPtr	-= len;
			return len;
		}

		len	= readPtr;
		readPtr	= 0;
		return 	len;
	}

	if (len < readPtr - writePtr)
	{
		readPtr	-= len;
		return len;
	}

	len	= readPtr - writePtr;
	readPtr	= writePtr;

	if (full)
		full	= 0;

	return len;
}

void Buffer::reset()
{
	full = readPtr	= writePtr	= 0;
}

int Buffer::resize(unsigned int newLen)
{
	return -1;
}

int Buffer::bufSuck(char *buf, unsigned int range)
{
	unsigned int	i;

	i	= bytesInBuffer();
	if (range < i)
		i	= range;

	if (i < (size - readPtr))
	{
		memcpy(buf, buffer + readPtr, i);
		readPtr	+= i;
	}
	else
	{
		int	j;

		j	= size - readPtr;
		memcpy(buf, buffer + readPtr, j);
		buf	+= j;
		j	= i - j;
		memcpy(buf, buffer, j);
		readPtr	= j;
	}

	if (	(full)	&& (i > 0)	)
		full	= 0;

	return i;
}

int Buffer::bufPeek(char *buf, unsigned int range)
{
	unsigned int	i;

	i	= bytesInBuffer();
	if (range < i)
		i	= range;

	if (i < (size - readPtr))
		memcpy(buf, buffer + readPtr, i);
	else
	{
		int	j;

		j	= size - readPtr;
		memcpy(buf, buffer + readPtr, j);
		buf	+= j;
		j	= i - j;
		memcpy(buf, buffer, j);
	}

	return i;
}

int Buffer::bufSend(Socket *sock)
{
	int		res;

	if (!bytesInBuffer())
		return 0;

	if (!(res = sock->testSend(0)))
		return 0;

	if (res == -1)
		return -1;

	if (writePtr > readPtr)
		res	= sock->callSend(buffer + readPtr, writePtr - readPtr);
	else
		res	= sock->callSend(buffer + readPtr, size - readPtr);

	if (res == -1)
		return -1;

	readPtr	+= res;
	if (readPtr >= size)
		readPtr	= 0;
	if ((full) && (res > 0))
		full 	= 0;

	return res;
}

/*
int Buffer::bufWrite(int fd)
{
	int	res;

	if (!bytesInBuffer())
		return 0;

	if (writePtr > readPtr)
	{
		res	= write(fd, buffer + readPtr, writePtr - readPtr);
		if (res == -1)
			return -1;

		readPtr	+= res;
		if (readPtr >= size)
			readPtr	-= size;
		if (full && (res > 0))
			full 	= 0;
		return res;
	}

	res	= write(fd, buffer + readPtr, size - readPtr);
	if (res == -1)
		return -1;

	readPtr	+= res;
	if (readPtr >= size)
		readPtr	= 0;
	if (	(full) && (res > 0)	)
		full 	= 0;

	return res;
}
*/
int Buffer::bufFWrite(FILE *f)
{
	int	res;

	if (!bytesInBuffer())
		return 0;

	if (writePtr > readPtr)
	{
		res	= fwrite(buffer + readPtr, 1, writePtr - readPtr, f);
		readPtr	= writePtr;
	}
	else
	{
		res	= fwrite(buffer + readPtr, 1, size - readPtr, f);
		readPtr	= 0;
	}
	if (readPtr >= size)
		readPtr	= 0;
	if (full && (res > 0))
		full 	= 0;

	return res;
}

int Buffer::bufRecv(Socket *sock)
{
	int		space;
	int		res;

	if (full)
		return 0;

	if (!((res = sock->testRecv(0))))
		return 0;

	if (writePtr >= readPtr)
		space	= size - writePtr;
	else
		space	= readPtr - writePtr;

	res	= sock->callRecv(buffer + writePtr, space);

	if (res > 0)
	{
		writePtr	+= res;
		if (writePtr >= size)
			writePtr	= 0;
		if (writePtr == readPtr)
			full		= 1;
		return res;
	}

	/*
	 * If we got here, 'select' says there's data and recv received
	 * either an error (-1) or 0 bytes. That last thing seems to
	 * mean the connection is closed, a -1 is returned to signal that.
	 */

	return -1;

}
/*
int Buffer::bufRead(int fd)
{
	int	space;
	int	res;

	if (full)
		return 0;

	if (writePtr >= readPtr)
		space	= size - writePtr;
	else
		space	= readPtr - writePtr;

	res	= read(fd, buffer + writePtr, space);
	if (res != -1)
	{
		writePtr	+= res;
		if (writePtr >= size)
			writePtr	= 0;
		if (	(writePtr == readPtr)	&&
			(res)	)
			full		= 1;
	}

	return res;
}
*/
int Buffer::bufFRead(FILE *f)
{
	int	space;
	int	res;

	if (full)
		return 0;

	if (writePtr >= readPtr)
		space	= size - writePtr;
	else
		space	= readPtr - writePtr;

	res	= fread(buffer + writePtr, 1, space, f);
	if (res != -1)
	{
		writePtr	+= res;
		if (writePtr >= size)
			writePtr	= 0;
		if (	(writePtr == readPtr)	&&
			(res)	)
			full		= 1;

	}

	return res;
}

unsigned int Buffer::bytesEmpty()
{
	if (full)
		return 0;

	if (writePtr >= readPtr)
		return size - readPtr + writePtr;

	return writePtr - readPtr;
}

unsigned int Buffer::bytesInBuffer()
{
	if (readPtr == writePtr)
	{
		if (full)
			return size;
		else
			return 0;
	}

	if (writePtr > readPtr)
		return writePtr - readPtr;

	return (writePtr + (size - readPtr));
}

unsigned int Buffer::getSize()
{
	return size;
}

/*
 * FAME
 */

char *Buffer::bufSuckp(unsigned int range)
{
	char	*buf;

	if (range > 0)
		if (bytesInBuffer() >= range)
		{
			if (!((buf	= (char *)malloc(range))))
			{
				fprintf(stderr, "Buffer : bufSuckp : malloc(%u) failed!!!\n", range);
				return 0;
			}
			if (bufSuck(buf, range) != (int)range)
			{
				free(buf);
				return 0;
			}
			return buf;
		}

	return 0;
}
