/***************************************************************************
 *                                                                         *
 *   This program is free software you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation either version 2 of the License, or      *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "Constants.h"
#include "DBElement.h"
#include "Socket.h"
#include "Buffer.h"
#include "GnutellaPacket.h"
#include "Packet.h"
#include "FSEntry.h"
#include "Path.h"
#include "File.h"

#include "CacheEntry.h"

CacheEntry::CacheEntry(char *newPattern, int newNumStrings)
{
	int	i;

	memcpy(pattern, newPattern, FILE_NAME_LENGTH);
	numStrings	= newNumStrings;
	if (numStrings > 1)
		sortPattern();
	response	= 0;

	i	= strlen(pattern);
	while (i < 4)
		pattern[i++]	= 0;
}

CacheEntry::CacheEntry(char *newPattern, Packet *newResponse,
				int newNumStrings)
{
	int	i;

	memcpy(pattern, newPattern, FILE_NAME_LENGTH);
	numStrings	= newNumStrings;
	if (numStrings > 1)
		sortPattern();
	response	= newResponse;

	i	= strlen(pattern);
	while (i < 4)
		pattern[i++]	= 0;
}

CacheEntry::~CacheEntry()
{
	if (response)
		delete response;
}

void CacheEntry::setResponse(Packet *newResponse)
{
	if (response)
		delete response;

	response	= newResponse;
}

/*
 * This function is the compare function for qsort
 * All it does is typecasting & dereferencing.
 */
int str_cmp(const void *arg1, const void *arg2)
{
	return (strcmp(*(char **)arg1, *(char **)arg2));
}

/*
 * This functions was written by Joris van Rantwijk because he couldn't stand
 * my own idea's about identifying cacheEntries based on a system of products
 * of prime numbers. He's right, this is better.
 */
void CacheEntry::sortPattern()
{
        char	*s;
	char	*p;
        int	i;
	char	**pList;

        s	= strdup(pattern);
        pList	= (char **)malloc(sizeof(char *) * numStrings);
        p	= s;

        for (i = 0; i < numStrings; i++)
	{
                p		= strtok(p, " ");
                pList[i]	= p;
		p		= 0;
	}

	/*
	 * Given the fact that sorting is a time-consuming
	 * business, I decided to do a simple swap when there are
	 * only 2 strings involved
	 */
	if (numStrings == 2)
	{
		char	*tmp;

		if (strcmp(pList[0], pList[1]) > 0)
		{
			tmp		= pList[0];
			pList[0]	= pList[1];
			pList[1]	= tmp;
		}
	}
	else
		qsort(pList, numStrings, sizeof(char *), str_cmp);
//	        quickSortPattern(pList, numStrings);

	p	= (char *)pattern;;
	*p	= 0;
	for (i = 0; i < numStrings; i++)
	{
		strcat(p, pList[i]);
	        p	+= strlen(p);
	        *p++	= ' ';
		*p	= 0;
	}
	*(p - 1)	= 0;

        free(s);
        free(pList);
}

int CacheEntry::getNumStrings()
{
	return numStrings;
}

/*
 * FAME
 */
char *CacheEntry::getPatternp()
{
	return (char *)pattern;
}

Packet *CacheEntry::getResponsep()
{
	return response;
}

/*
 * DBElement inheritance
 */
int CacheEntry::equals(DBElement *object)
{
	return (!strncmp(pattern, ((CacheEntry *)object)->getPatternp(),
				FILE_NAME_LENGTH));
}

int CacheEntry::lessThan(DBElement *object)
{
	return (strncmp(pattern, ((CacheEntry *)object)->getPatternp(),
				FILE_NAME_LENGTH) < 0);
}

int CacheEntry::greaterThan(DBElement *object)
{
	return (strncmp(pattern, ((CacheEntry *)object)->getPatternp(),
				FILE_NAME_LENGTH) > 0);
}

unsigned int CacheEntry::hashValue()
{
	return (*(unsigned int *)pattern);
}
