/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
 
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>

#include "Config.h"
#include "Constants.h"
#include "ControlCentre.h"
#include "DBElement.h"
#include "DB.h"
#include "GnutellaPacket.h"
#include "SearchResult.h"
#include "Packet.h"
#include "Mesg.h"
#include "Socket.h"
#include "Buffer.h"
#include "Connection.h"
#include "GnutellaConnection.h"
#include "Host.h"
#include "FSEntry.h"
#include "Path.h"
#include "File.h"
#include "Inet.h"
#include "Transfer.h"
#include "Upload.h"
#include "Download.h"
#include "TransferStatus.h"
#include "LocalResult.h"
#include "HTTPServer.h"
#include "FSScanner.h"
#include "FServer.h"
#include "Query.h"
#include "Monitor.h"
#include "HTTPClient.h"
#include "Downloader.h"
#include "SearchManager.h"
#include "Networker.h"
#include "HostCatcher.h"
#include "GnutellaRouter.h"
#include "Backend.h"
#include "Commander.h"

Commander::Commander(ControlCentre *newMCC)
{
	MCC			= newMCC;
	running			= 1;
	prevCommand[0]		= 0;
	prevListCommand[0]	= 0;
	os			= 0;
}

int Commander::keepRunning()
{
	return running;
}

int Commander::doCommand(char *buf, Backend *b, char *outbuf)
{
	int		cmd;
	char		*chr;

	outbuf[0]	= 0;

	while (*buf == ' ')
		buf++;

	if ( (buf[0] == '\r') || (buf[0] == '\n') )
		return 1;

	if (buf[0] == '!')
		buf	= &prevCommand[0];
	else
		if (buf[0] != 'c')
			strncpy(prevCommand, buf, COMMAND_BUF_SIZE);

	if (((cmd	= interpCommand(buf))) == COMMAND_LIST)
	{
		if (*buf == 'c')
		{
			chr	= (char *)prevListCommand;
			while (*chr == ' ')
				chr++;

			if (*chr != 'l')
			{
				sprintf(outbuf, "Error : continue "
					"requested without prior list command!!!\n");
				return 0;
			}
			while ((*chr != ' ') && (*chr))
				chr++;
			while ((*chr == ' ') && (*chr))
				chr++;

			if (*(chr + 1))
				sprintf(buf, "l %c%c %d", *chr, *(chr + 1), os);
			else
				sprintf(buf, "l %c %d", *chr, os);
		}
		else 
			if (*buf == 'l')
			{
				char	*offset;

				for (offset = buf; *offset && ( (*offset < '0') || (*offset > '9') ); 
								offset++)
					;
				os	= atoi(offset);
				strncpy(prevListCommand, buf, COMMAND_BUF_SIZE);
			}
	}

	/*
	 * fist, skip command; then skip spaces
	 */
	while ((*buf != ' ') && (*buf))
		buf++;
	while ((*buf == ' ') && (*buf))
		buf++;
	switch (cmd)
	{
		case COMMAND_HOST	: doHost(buf, b, outbuf);
					break;
		case COMMAND_HELP	: doHelp(buf, b, outbuf);
					break;
		case COMMAND_HOLDON	: doHoldon(buf, b, outbuf);
					break;
		case COMMAND_SET	: doSet(buf, b, outbuf);
					break;
		case COMMAND_SEARCH	: doSearch(buf, b, outbuf);
					break;
		case COMMAND_SHARE	: doShare(buf, b, outbuf);
					break;
		case COMMAND_SCAN	: doScan(buf, b, outbuf);
					break;
		case COMMAND_SAVE	: doSave(buf, b, outbuf);
					break;
		case COMMAND_LIST	:
					doList(buf, b, outbuf);
					break;
		case COMMAND_DOWNLOAD	: doDownload(buf, b, outbuf);
					break;
		case COMMAND_KILL	: doKill(buf, b, outbuf);
					break;
		case COMMAND_VERSION	: doVersion(buf, b, outbuf);
					break;
		case COMMAND_QUIT	: doQuit(buf, b, outbuf);
					break;
/*
		case COMMAND_	: do(buf, b, outbuf);
					break;
*/
		case COMMAND_INVALID	:
					MCC->addError("Illegal command.");
					break;
		default	: MCC->addError("BUG : Don't know how to handle command");
					break;
	}

	return 1;
}

int Commander::doVersion(char *buf, Backend *b, char *outbuf)
{
	sprintf(outbuf, "%s\r\n", VERSION);
	return 1;
}


int Commander::doHoldon(char *buf, Backend *b, char *outbuf)
{
	strcpy(outbuf, "'holdon' is only available in DAEMON mode and must be"
			"spelled out");
	return 1;
}

int Commander::doQuit(char *buf, Backend *b, char *outbuf)
{
	char    fn[FILE_NAME_LENGTH];

	if (MCC->get_AUTO_SAVE_HOSTS())
	{
		sprintf(fn, "%s/.hagelslag/hagelslag.hosts", getenv("HOME"));
		if (b->writeHosts(fn) == -1)
                        MCC->addError("couldn't save hosts!");
	}
	running	= 0;
	return 1;
}

int Commander::doHost(char *buf, Backend *b, char *outbuf)
{
	int		res;
	int		p;
	int		i;
	char		dotted[DOTTED_LENGTH];

	p	= buf[0];
	while ((*buf != ' ') && (*buf))
		buf++;
	while ((*buf == ' ') && (*buf))
		buf++;
	switch (p)
	{
		case	'a'	:
				for (i = 0; buf[i]; ++i)
				{
					char	*c;

					if (!strchr("0123456789.:", buf[i]))
					{
						if ((c = strchr(buf, ':')))
							*c	= 0;
						
						if (c)
							Inet::IPPortToDotted(dotted, Inet::NSLookup(buf), atoi(++c));
						else
							Inet::IPPortToDotted(dotted, Inet::NSLookup(buf), 6346);

						if (((res = b->storeHost(dotted, 0, 0))) == ENTRY_INVALID)
						{
							sprintf(outbuf, "Add host %s FAILED!\r\n", buf);
							return 0;
						}
						else
							sprintf(outbuf, "Host added at entry %d\r\n", res);
								return 1;
					}
				}
				if (!strchr(buf, ':'))
					sprintf(buf, "%s:6346", buf);
				if (((res = b->storeHost(buf, 0, 0))) == ENTRY_INVALID)
				{
					sprintf(outbuf, "Add host %s FAILED!\r\n", buf);
					return 0;
				}
				else
					sprintf(outbuf, "Host added at entry %d\r\n", res);
				break;
		case	'd'	:
				if (b->deleteHost(atoi(buf)) == ENTRY_INVALID)
				{
					sprintf(outbuf, "Delete host %d FAILED!\r\n", atoi(buf));
					return 0;
				}
				break;
		case	'c'	:
				if (b->connectToHost(atoi(buf)) == ENTRY_INVALID)
				{
					sprintf(outbuf, "Connect to %s FAILED!\r\n", buf);
					return 0;
				}
				break;
		default	:
				sprintf(outbuf, "add, del or connect expected\r\n");
				return 0;
	}

	return 1;
}

int Commander::doSet(char *buf, Backend *b, char *outbuf)
{
	outbuf[0]	= 0;

	if (!strncmp("MAX_TTL", buf, strlen("MAX_TTL")))
	{
		int	val;

		buf	+= strlen("MAX_TTL");
		while(*buf == ' ')
			buf++;

		val	= atoi(buf);
		if ( (val >= 0) && (val <= 255 ) )
			b->set_INT_PARAM(PARAM_MAX_TTL, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("DEFAULT_TTL", buf, strlen("DEFAULT_TTL")))
	{
		int	val;

		buf	+= strlen("DEFAULT_TTL");
		while(*buf == ' ')
			buf++;

		val	= atoi(buf);
		if ( (val >= 0) && (val <= 255) )
			b->set_INT_PARAM(PARAM_DEFAULT_TTL, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("ROUTING_TABLE_SIZE", buf,
					strlen("ROUTING_TABLE_SIZE")))
	{
		int	val;

		buf	+= strlen("ROUTING_TABLE_SIZE");
		while(*buf == ' ')
			buf++;

		val	= atoi(buf);
		if ( (val > 0) && (val < 100000) )
			b->set_INT_PARAM(PARAM_ROUTING_TABLE_SIZE, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("MAX_CONNECTIONS", buf, strlen("MAX_CONNECTIONS")))
	{
		int	val;

		buf	+= strlen("MAX_CONNECTIONS");
		while(*buf == ' ')
			buf++;

		val	= atoi(buf);
		if ( (val >= 0) && (val < 20000) )
			b->set_INT_PARAM(PARAM_MAX_CONNECTIONS, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("MAX_RESULTS", buf, strlen("MAX_RESULTS")))
	{
		int	val;

		buf	+= strlen("MAX_RESULTS");
		while(*buf == ' ')
			buf++;

		val	= atoi(buf);
		if ( (val > 0) && (val < 15000) )
			b->set_INT_PARAM(PARAM_MAX_RESULTS, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("MAX_HOPS", buf, strlen("MAX_HOPS")))
	{
		int	val;

		buf	+= strlen("MAX_HOPS");
		while(*buf == ' ')
			buf++;

		val	= atoi(buf);
		if ( (val >= 0) && (val <= 255) )
			b->set_INT_PARAM(PARAM_MAX_HOPS, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("MAX_UPLOADS", buf, strlen("MAX_UPLOADS")))
	{
		int	val;

		buf	+= strlen("MAX_UPLOADS");
		while(*buf == ' ')
			buf++;

		val	= atoi(buf);
		if ( (val >= 0) && (val <= 5000) )
			b->set_INT_PARAM(PARAM_MAX_UPLOADS, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("MAX_DOWNLOADS", buf, strlen("MAX_DOWNLOADS")))
	{
		int	val;

		buf	+= strlen("MAX_DOWNLOADS");
		while(*buf == ' ')
			buf++;

		val	= atoi(buf);
		if ( (val >= 0) && (val <= 5000) )
			b->set_INT_PARAM(PARAM_MAX_DOWNLOADS, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("PORT", buf, strlen("PORT")))
	{
		int	val;

		buf	+= strlen("PORT");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if ( (val >= 0) && (val <= 65535) )
			b->set_INT_PARAM(PARAM_PORT, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("SPEED", buf, strlen("SPEED")))
	{
		int	val;

		buf	+= strlen("SPEED");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if (val >= 0)
			b->set_INT_PARAM(PARAM_SPEED, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("DOWNLOAD_DIR", buf, strlen("DOWNLOAD_DIR")))
	{
		buf	+= strlen("DOWNLOAD_DIR");

		while(*buf == ' ')
			buf++;
		b->set_STR_PARAM(PARAM_DOWNLOAD_DIR, buf);
	}
	else if (!strncmp("MAX_MONITOR_LENGTH", buf, 
					strlen("MAX_MONITOR_LENGTH")))
	{
		int	val;

		buf	+= strlen("MAX_MONITOR_LENGTH");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if ( (val >= 0) && (val <= 1000 ) )
			b->set_INT_PARAM(PARAM_MAX_MONITOR_LENGTH, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("AUTO_PUSH", buf, strlen("AUTO_PUSH")))
	{
		int	val;

		buf	+= strlen("AUTO_PUSH");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		b->set_INT_PARAM(PARAM_AUTO_PUSH, val);
	}
	else if (!strncmp("AUTO_CLEAR_COMPLETE", buf, 
					strlen("AUTO_CLEAR_COMPLETE")))
	{
		int	val;

		buf	+= strlen("AUTO_CLEAR_COMPLETE");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		b->set_INT_PARAM(PARAM_AUTO_CLEAR_COMPLETE, val);
	}
	else if (!strncmp("AUTO_CONNECTIONS", buf, strlen("AUTO_CONECTIONS")))
	{
		int	val;

		buf	+= strlen("AUTO_CONNECTIONS");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if ((val >= 0) && (val <= MCC->get_MAX_CONNECTIONS()) )
			b->set_INT_PARAM(PARAM_AUTO_CONNECTIONS, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("AUTO_SAVE_HOSTS", buf, strlen("AUTO_SAVE_HOSTS")))
	{
		int	val;

		buf	+= strlen("AUTO_SAVE_HOSTS");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		b->set_INT_PARAM(PARAM_AUTO_SAVE_HOSTS, val);
	}
	else if (!strncmp("AUTO_CLEAR_RESULTS", buf, 
					strlen("AUTO_CLEAR_RESULTS")))
	{
		int	val;

		buf	+= strlen("AUTO_CLEAR_RESULTS");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		b->set_INT_PARAM(PARAM_AUTO_CLEAR_RESULTS, val);
	}
	else if (!strncmp("MAX_HOST_CON_FAIL", buf, 
					strlen("MAX_HOST_CON_FAIL")))
	{
		int	val;

		buf	+= strlen("MAX_HOST_CON_FAIL");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if (val >= 0)
			b->set_INT_PARAM(PARAM_MAX_HOST_CON_FAIL, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("MAX_FILES_SHARED", buf, strlen("MAX_FILES_SHARED")))
	{
		int	val;

		buf	+= strlen("MAX_FILES_SHARED");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if ( (val >= 0) && (val <= FS_MAX) )
			b->set_INT_PARAM(PARAM_MAX_FILES_SHARED, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("MAX_CON_ERRORS", buf, strlen("MAX_CON_ERRORS")))
	{
		int	val;

		buf	+= strlen("MAX_CON_ERRORS");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if (val >= 0)
			b->set_INT_PARAM(PARAM_MAX_CON_ERRORS, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("IGNORE_LOCAL_IP", buf, strlen("IGNORE_LOCAL_IP")))
	{
		int	val;

		buf	+= strlen("IGNORE_LOCAL_IP");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if (val >= 0)
			b->set_INT_PARAM(PARAM_IGNORE_LOCAL_IP, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("NOTIFY_RESULTS", buf, strlen("NOTIFY_RESULTS")))
	{
		int	val;

		buf	+= strlen("NOTIFY_RESULTS");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if (val >= 0)
			b->set_INT_PARAM(PARAM_NOTIFY_RESULTS, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("CACHE_ENTRIES", buf, strlen("CACHE_ENTRIES")))
	{
		int	val;

		buf	+= strlen("CACHE_ENTRIES");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if (val >= 0)
			b->set_INT_PARAM(PARAM_CACHE_ENTRIES, val);
		else
			MCC->addError("Invalid parameter.");
	}
	else if (!strncmp("PACKET_RATIO", buf, strlen("PACKET_RATIO")))
	{
		int	val;

		buf	+= strlen("PACKET_RATIO");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if ( (val >= 0) && (val < 500000) )
			b->set_INT_PARAM(PARAM_PACKET_RATIO, val);
		else
			MCC->addError("Invalid parameter.");
	}
/*
	else if (!strncmp("", buf, strlen("")))
	{
		int	val;

		buf	+= strlen("");

		while(*buf == ' ')
			buf++;
		val	= atoi(buf);
		if ( (val > ) && (val < ) )
			b->set__PARAM(PARAM_, val);
		else
			MCC->addError("Invalid parameter.");
	}
*/
	else
		MCC->addError("Setting does not exist.");

	return 1;
}

int Commander::doHelp(char *buf, Backend *b, char *outbuf)
{
	if (!strncmp("host", buf, strlen("host")))
	{
		sprintf(outbuf, "host add <1.2.3.4:5>\r\n"
				"Add host with ip address 1.2.3.4 and port 5.\r\n\r\n"
				"host <del|connect> <n>\r\n"
				"Delete hosts number n or connect to host "
				"number n.\r\nSee also \"list\".\r\n");
	}
	else if (!strncmp("set", buf, strlen("set")))
	{
		sprintf(outbuf, "set <parameter> <n>\r\n"
				"Set a parameter to a value n. Possible "
				"parameters are : \r\n"
				"MAX_TTL DEFAULT_TTL MAX_CONNECTIONS "
				"ROUTING_TABLE_SIZE MAX_HOPS MAX_UPLOADS\r\n"
				"MAX_DOWNLOADS MAX_RESULTS PORT "
				"SPEED DOWNLOAD_DIR\r\nAUTO_CLEAR_COMPLETE "
				"AUTO_PUSH AUTO_CONNECTIONS "
				"MAX_MONITOR_LENGTH.\r\n"
				"AUTO_SAVE_HOSTS AUTO_CLEAR_RESULTS"
				"MAX_FILES_SHARED MAX_CON_ERRORS\r\n"
				"IGNORE_LOCAL_IP NOTIFY_RESULTS CACHE_ENTRIES PACKET_RATIO\r\n");
	}
	else if (!strncmp("list", buf, strlen("list")))
	{
		sprintf(outbuf, "list <results|downloads|uploads|"
				"connections|shares|hosts|options|stats"
				"|queries|internals>\r\n"
				"Shows the list of requested objects.\r\n"
			"NOTE : the hosts, results and connections arguments "
				"require an extra parameter \r\n<n> which"
				"is an offset after which several elements "
				"are listed.\r\n");
	}
	else if (!strncmp("download", buf, strlen("download")))
	{
		sprintf(outbuf, "download <start|continue|push> <n>\r\n"
				"Download result with index \"n\". "
				"\"n\" can be found with \"list\".\r\n"
				"\"start\" will delete the file if it exist and "
				"start from the beginning.\r\n\"continue\" will "
				"try to continue the download. push will "
				"try\r\nto start the download with a push "
				"request.\r\n");
	}
	else if (!strncmp("search", buf, strlen("search")))
	{
		sprintf(outbuf, "search <speed> <pattern>\r\n"
				"search clear\r\n"
				"Either search \"pattern\" from hosts with "
				"speed greater or equal to \"speed\", "
				"or\r\nclear all search results.\r\n");
	}
	else if (!strncmp("version", buf, strlen("version")))
	{
		sprintf(outbuf, "version\r\n"
				"Prints a version string.\r\n");
	}
	else if (!strncmp("share", buf, strlen("share")))
	{
		sprintf(outbuf, "share <add|del|clear> <path>\r\n"
				"Add or delete one directory from the list "
				"of shared directories.\r\nAll shared dir"
				"ectories can be removed from the list "
				"with \"clear\" is used.\r\n"
				"See also \"scan\".\r\n");
	}
	else if (!strncmp("kill", buf, strlen("kill")))
	{
		sprintf(outbuf, "kill <upload|download|connection> <n>\r\n"
				"terminate the upload or download number n. "
				"\"n\" can be found with \"list\".\r\n");
	}
	else if (!strncmp("scan", buf, strlen("scan")))
	{
		sprintf(outbuf, "scan\r\n"
				"Scan the shared directories. This must "
				"be done after each call to \"share\".\r\n");
	}
	else if (!strncmp("save", buf, strlen("save")))
	{
		sprintf(outbuf, "save <hosts|options|shares>\r\n"
				"Save the host database, the current "
				"options or the list of shares to file.\r\n");
	}
	else if (!strncmp("continue", buf, strlen("continue")))
	{
		sprintf(outbuf, "continue\r\nScroll new data of last list command.\r\n");
	}
	else if (!strncmp("quit", buf, strlen("quit")))
	{
		sprintf(outbuf, "quit\r\nQuit the program. Jezus, what else "
				"did you expect ?\r\n");
	}
#ifdef DAEMON
	else if (!strncmp("holdon", buf, strlen("holdon")))
	{
		sprintf(outbuf, "holdon\r\n"
				"Only available when running as deamon. "
				"\"holdon\" drops the connection,\r\n"
				"but the deamon will continue to run. "
				"Use \"quit\" to really stop the deamon.\r\n");
	}
#endif
	else
	{
		sprintf(outbuf, "help		: display help.\r\n"
				"help <command>	: display specific help.\r\n"
				"where \"command\" is one of the following :\r\n"
				"host set help list download search kill version "
				"share scan save quit continue"
#ifdef DAEMON
				"holdon"
#endif
				"\r\nMost command and some parameters can be abbreviated, "
				"but not all.\r\n");
	}

/*
	else if (!strncmp("", buf, strlen("")))
	{
		sprintf(outbuf, "\r\n");
	}
*/

	return 1;
}

int Commander::doSearch(char *buf, Backend *b, char *outbuf)
{
	int	spd;

	if (!strncmp(buf, "clear", strlen("clear")))
	{
		b->clearResults();
		return 1;
	}

	spd	= atoi(buf);

	if (spd < 0)
	{
		MCC->addError("negative speed.");
		return 0;
	}

	while (strchr("0123456789", *buf))
		buf++;
	while(*buf == ' ')
		buf++;

	return b->launchQuery(spd, buf);
}

int Commander::doList(char *buf, Backend *b, char *outbuf)
{
	TransferStatus		*t, *prv;
	SearchResult		*s;
	GnutellaConnection	*c;
	Host			*h;
	Path			*p;
	char			peer[DOTTED_LENGTH];
	char			st[STATE_NAME_LENGTH];
	char			fn[FILE_NAME_LENGTH];
	int			i;
	int			pm; // first letter of parameter
	int			pp; // 2nd letter of parameter
	int			ratio;
	int			index;

	pm	= buf[0];
	pp	= buf[1];

	while ((*buf != ' ') && (*buf))
		buf++;
	while ((*buf == ' ') && (*buf))
		buf++;

	switch(pm)
	{
		case	'u'	:
			t	= b->getUploads();

			while (t)
			{
				int	tm;

				getStateName(t->getState(), st);

				tm	= (t->getStampNow() - t->getStamp());
				if (tm < 1)
					tm	= 1;
				ratio	= 	   t->getBytesDone() 		/
					 //------------------------------------
							tm;
				t->getFileName(fn);
				mkPrintable(fn);
				sprintf(outbuf, "%d\t: %s ---> %s\r\n"
						"\tState: %s\tFileID: %d\t"
						"%d / %d bytes %d.%d KB/s.\r\n",
						t->getTransferID(), fn,
						t->getPeerDottedp(), st, t->getFileIndex(),
						t->getBytesTotal() - t->getBytesLeft(),
						t->getBytesTotal(), ratio / 1024,
						ratio % 1024);

				outbuf	+= strlen(outbuf);
				prv	= t;
				t	= t->getNext();
				delete prv;
			}
			break;
		case	'd'	:
			t	= b->getDownloads();

			while (t)
			{
				int	tm;

				getStateName(t->getState(), st);

				tm	= (t->getStampNow() - t->getStamp());
				if (tm < 1)
					tm	= 1;
				ratio	= 	t->getBytesDone()	 	/
					 //------------------------------------
							tm;

				t->getFileName(fn);
				mkPrintable(fn);
				sprintf(outbuf, "%d\t: %s ---> %s\r\n"
						"\tState: %s\tFileID: %d\t"
						"%d / %d bytes %d.%d KB/s.\r\n",
						t->getTransferID(), t->getPeerDottedp(),
						fn, st, t->getFileIndex(),
						t->getBytesTotal() - t->getBytesLeft(),
						t->getBytesTotal(), ratio / 1024,
						ratio % 1024);

				outbuf	+= strlen(outbuf);
				prv	= t;
				t	= t->getNext();
				delete prv;
			}
			break;
		case	'c'	:
			index	= atoi(buf);
			if (index >= MCC->get_MAX_CONNECTIONS())
				index	= MCC->get_MAX_CONNECTIONS() - 11;
			if (index < 0)
				index	= 0;
			for (i = index; ( (i < MCC->get_MAX_CONNECTIONS()) &&
					(i < index + 10) ); i++)
			{
				if (!((c = b->getConnection(i))))
					continue;
				if (!((h = b->getHost(c->getHostIndex()))))
					continue;

				h->getDotted(peer);
				getStateName(c->getState(), st);
				sprintf(outbuf, "%d : %s\tSTATE: %s\tCID: %s\t"
						"OUTGOING: %s\r\n\tSENT: %d\t"
						"RECEIVED: %d\tDROPPED: %d\t"
						"ERRORS: %d\r\n",
						i, peer, st,
						h->hasClientID() ? "YES":"NO",
						c->isOutgoing() ? "YES" : "NO",
						c->getSent(), c->getRecvd(),
						c->getDropped(), c->getErrors());
				outbuf	+= strlen(outbuf);
			}
			os	+= 10;
			break;
		case	's'	:
			if (pp == 't')
			{
				sprintf(outbuf, "Hosts: %u\tFiles On Net: %lu\t"
						"Size on Net: %lu\tLocal searches %u\r\n"
						"Files Shared: %d\tSize Shared: %d\r\n",
						b->getHostsOnNet(), b->getFilesOnNet(),
						b->getSizeOnNet(), b->getLocalSearches(),
						b->getFilesShared(), b->getSizeShared());
				break;
			}
			for (p = b->getPathList(); p; p = (Path *)p->getNext())
			{
				if (!p->isSubdir())
				{
					sprintf(outbuf, "%s\r\n", p->getPathp());
					outbuf	+= strlen(outbuf);
				}
			}
			break;
		case	'h'	:
			index	= atoi(buf);
			if (index >= MCC->get_MAX_HOSTS())
				index	= MCC->get_MAX_HOSTS() - 21;
			if (index < 0)
				index	= 0;
			for (i = index; ( (i < MCC->get_MAX_HOSTS()) &&
					(i < index + 20) ); i++)
			{
				if (!((h = b->getHost(i))))
					continue;

				h->getDotted(peer);
				sprintf(outbuf, "%d\tHost: %s\tREAL: %s\t"
					"CID: %s\tTries: %d\r\n", i, peer,
					h->isReal() ? "YES" : "NO",
					h->hasClientID() ? "YES" : "NO",
					h->getTries());
				outbuf	+= strlen(outbuf);
			}
			os	+= i - index;
			break;
		case	'r'	:
			int		i;

			index	= atoi(buf);
			if (index >= MCC->get_MAX_RESULTS())
				index	= MCC->get_MAX_RESULTS() - 11;
			if (index < 0)
				index	= 0;
			for (i = index; ( (i < MCC->get_MAX_RESULTS()) &&
					(i < index + 10) ); i++)
				if ((s = b->getOneResult(i)))
				{
					char	ip[25];
					
					Inet::IPToDotted(ip, s->getHostIP());
					s->getName(fn);
					mkPrintable(fn);
					sprintf(outbuf, "%d :\t%s\r\n\tFileIndex: "
							"%d\tSize: %d\tHost: %s\t"
							"Speed: %d.\r\n",i, fn,
							s->getFileIndex(),
							s->getSize(), ip,
							s->getHostSpeed());
					outbuf	+= strlen(outbuf);
				}
			os	+= i - index;
			break;
		case	'o'	:
			char	downdir[FILE_NAME_LENGTH];

			MCC->get_DOWNLOAD_DIR(downdir);
			for (i = 0; i < PARAM_INVALID; ++i)
			{
				switch(i)
				{
					case	PARAM_MAX_TTL	:
						sprintf(outbuf, "MAX_TTL %d\r\n",
							MCC->get_MAX_TTL());
						break;
					case	PARAM_DEFAULT_TTL	:
						sprintf(outbuf, "DEFAULT_TTL %d\r\n",
							MCC->get_DEFAULT_TTL());
						break;
					case	PARAM_ROUTING_TABLE_SIZE	:
						sprintf(outbuf, "ROUTING_TABLE_SIZE %d\r\n",
							MCC->get_ROUTING_TABLE_SIZE());
						break;
					case	PARAM_MAX_CONNECTIONS	:
						sprintf(outbuf, "MAX_CONNECTIONS %d\r\n",
							MCC->get_MAX_CONNECTIONS());
						break;
					case	PARAM_MAX_RESULTS	:
						sprintf(outbuf, "MAX_RESULTS %d\r\n",
							MCC->get_MAX_RESULTS());
						break;
					case	PARAM_MAX_HOPS	:
						sprintf(outbuf, "MAX_HOPS %d\r\n",
							MCC->get_MAX_HOPS());
						break;
					case	PARAM_MAX_PUSH_REQUESTS	:
						sprintf(outbuf, "MAX_PUSH_REQUESTS %d\r\n",
							MCC->get_MAX_PUSH_REQUESTS());
						break;
					case	PARAM_MAX_UPLOADS	:
						sprintf(outbuf, "MAX_UPLOADS %d\r\n",
							MCC->get_MAX_UPLOADS());
						break;
					case	PARAM_MAX_DOWNLOADS	:
							sprintf(outbuf, "MAX_DOWNLOADS %d\r\n",
							MCC->get_MAX_DOWNLOADS());
						break;
					case	PARAM_MAX_HOSTS	:
						sprintf(outbuf, "MAX_HOSTS %d\r\n",
							MCC->get_MAX_HOSTS());
						break;
					case	PARAM_PORT	:
						sprintf(outbuf, "PORT %d\r\n",
							MCC->get_PORT());
						break;
					case	PARAM_SPEED	:
						sprintf(outbuf, "SPEED %d\r\n",
							MCC->get_SPEED());
						break;
					case	PARAM_MAX_MONITOR_LENGTH	:
						sprintf(outbuf, "MAX_MONITOR_LENGTH %d\r\n",
							MCC->get_MAX_MONITOR_LENGTH());
						break;
					case	PARAM_AUTO_CLEAR_COMPLETE	:
						sprintf(outbuf, "AUTO_CLEAR_COMPLETE %d\r\n",
							MCC->get_AUTO_CLEAR_COMPLETE());
						break;
					case	PARAM_AUTO_CONNECTIONS	:
						sprintf(outbuf, "AUTO_CONNECTIONS %d\r\n",
							MCC->get_AUTO_CONNECTIONS());
						break;
					case	PARAM_AUTO_PUSH	:
						sprintf(outbuf, "AUTO_PUSH %d\r\n",
							MCC->get_AUTO_PUSH());
						break;
					case	PARAM_AUTO_CLEAR_RESULTS	:
						sprintf(outbuf, "AUTO_CLEAR_RESULTS %d\r\n",
							MCC->get_AUTO_CLEAR_RESULTS());
						break;
					case	PARAM_MAX_HOST_CON_FAIL	:
						sprintf(outbuf, "MAX_HOST_CON_FAIL %d\r\n",
							MCC->get_MAX_HOST_CON_FAIL());
						break;
					case	PARAM_AUTO_SAVE_HOSTS	:
						sprintf(outbuf, "AUTO_SAVE_HOSTS %d\r\n",
							MCC->get_AUTO_SAVE_HOSTS());
						break;
					case	PARAM_MAX_FILES_SHARED	:
						sprintf(outbuf, "MAX_FILES_SHARED %d\r\n",
							MCC->get_MAX_FILES_SHARED());
						break;
					case	PARAM_MAX_CON_ERRORS	:
						sprintf(outbuf, "MAX_CON_ERRORS %d\r\n",
							MCC->get_MAX_CON_ERRORS());
						break;
					case	PARAM_IGNORE_LOCAL_IP	:
						sprintf(outbuf, "IGNORE_LOCAL_IP %d\r\n",
							MCC->get_IGNORE_LOCAL_IP());
						break;
					case	PARAM_NOTIFY_RESULTS	:
						sprintf(outbuf, "NOTIFY_RESULTS %d\r\n",
							MCC->get_NOTIFY_RESULTS());
						break;
					case	PARAM_CACHE_ENTRIES	:
						sprintf(outbuf, "CACHE_ENTRIES %d\r\n",
							MCC->get_CACHE_ENTRIES());
						break;
					case	PARAM_PACKET_RATIO	:
						sprintf(outbuf, "PACKET_RATIO %d\r\n",
							MCC->get_PACKET_RATIO());
						break;
					case	PARAM_DAEMON_PORT	:
#ifdef DAEMON
						sprintf(outbuf, "DAEMON_PORT %d\r\n",
							MCC->get_DAEMON_PORT());
#endif
						break;
					case	PARAM_DOWNLOAD_DIR	:
						sprintf(outbuf, "DOWNLOAD_DIR ");
						outbuf	+= strlen(outbuf);
						MCC->get_DOWNLOAD_DIR(outbuf);
						outbuf	+= strlen(outbuf);
						sprintf(outbuf, "\r\n");
						break;
					case	PARAM_CLIENTID	:
						break;
/*
					case	PARAM_	:
						sprintf(outbuf, " %d\r\n",
							MCC->get_());
						break;
*/
					default	: MCC->addError("BUG : doList options : Unknown parameter !!!");
				}
				outbuf	+= strlen(outbuf);
			}
			break;
		case	'q'	:
			Query	*q;
			int	speed;

			for (q = b->getQueries(); q; q = q->getNext())
			{
				speed	= q->getSpeed();
				// actually, fn is not a filename here but who cares ?
				q->getQuery(fn);
				mkPrintable(fn);
				sprintf(outbuf, "%d %s\r\n", speed, fn);
				outbuf	+= strlen(outbuf);
			}
			break;
		case	'i'	:
				sprintf(outbuf,
						/* FileServer */
						"FileServer :\r\n"
						"\tMemory usage :\r\n"
						"\t\tSearch Array\t %u\r\n"
						"\t\tHashArray\t %u\r\n"
						"\t\tTotal :\t\t %u\r\n"
						"\tStats :\r\n"
						"\t\tFiles Shared :\t %u\r\n"
						"\t\tCache Size :\t %u\r\n"
						"\t\tCache Entries :\t %u\t(%u %%)\r\n"
						"\t\tLocal Searches : %u\r\n"
						"\t\tCache Hits :\t %u\t(%u %%)\r\n"
						"\r\n"

						// Downloader
						"Downloader :\r\n"
						"\tMemory usage :\r\n"
						"\tStats :\r\n"
						"\t\tBytes downloaded :\t %u MB\r\n"
						"\r\n"

						// HostCatcher
						"HostCatcher :\r\n"
						"\tMemory usage :\r\n"
						"\tStats :\r\n"
						"\t\tHosts added :\t %u\r\n"
						"\r\n"

						// GnutellaRouter
						"Gnutella Router :\r\n"
						"\tMemory usage :\r\n"
						"\tStats :\r\n"
						"\t\tInit Packets :\t\t %u\r\n"
						"\t\tInit Resp. Packets :\t %u\r\n"
						"\t\tPush Packets :\t\t %u\r\n"
						"\t\tSearch Packets :\t %u\r\n"
						"\t\tSearch Resp Packets :\t %u\r\n"
						"\t\tTotal Packets :\t\t %u\r\n"
						"\t\tDouble Packets :\t %u\r\n"
						"\t\tResponse Found :\t %u\r\n"
						"\t\tResponses Missed :\t %u\r\n"
						"\t\tMissed / Found Ratio :\t %u %%\r\n"
						"\r\n"

						// SearchManager
						"Search Manager :\r\n"
						"\tMemory usage :\r\n"
						"\tStats :\r\n"
						"\t\tSearches Started :\t %u\r\n"
						"\t\tResults Found :\t\t %u\r\n"
						"\t\tAvg Results / Search :\t %u\r\n"
						"\r\n"

						// Monitor
						"Monitor :\r\n"
						"\tMemory usage :\r\n"
						"\tStats :\r\n"
						"\t\tQueries :\t %u\r\n"
						"\r\n"

						// Networker
						"Networker :\r\n"
						"\tMemory usage :\r\n"
						"\tStats :\r\n"
						"\t\tBytes Incoming :\t %u MB\r\n"
						"\t\tBytes Outgoing :\t %u MB\r\n"
						"\r\n"

/*
						//
						" :\r\n"
						"\tMemory usage :\r\n"
						"\tStats :\r\n"
						"\t\t :\t %u\r\n"
						"\r\n"

*/
						// FileServer
						, b->getSearchArraySize(), b->getHashArraySize(),
						b->getSearchArraySize() + b->getHashArraySize(),
						b->getFilesShared(), b->getCacheSize(),
						b->getCacheEntries(),
						(100 * b->getCacheEntries()) / (b->getCacheSize() | 0x1),
						b->getLocalSearches(), b->getCacheHits(),
						(100 * b->getCacheHits()) / (b->getLocalSearches() | 0x1),

						// Downloader
						(b->getDLBytesDownloaded() / (1024 * 1024)),

						// HostCatcher
						b->getHCHostsCaught(),

						// Gnutella Router
						b->getGRInitPackets(),
						b->getGRInitRespPackets(),
						b->getGRPushPackets(),
						b->getGRSearchPackets(),
						b->getGRSearchRespPackets(),
						(
						b->getGRInitPackets()		+
						b->getGRInitRespPackets()	+
						b->getGRPushPackets()		+
						b->getGRSearchPackets()		+
						b->getGRSearchRespPackets()
						),
						b->getGRDoublePackets(),
						b->getGRRespMatches(),
						b->getGRRespMisses(),
						(
						(b->getGRRespMisses() * 100)	/
						//------------------------------
						(b->getGRRespMatches() | 0x1)
						),

						// SearchManager
						b->getSMQueries(),
						b->getSMResultCount(),
						(
						b->getSMResultCount()	/
						//-------------------
						(b->getSMQueries() | 0x1)
						),

						// Monitor
						b->getQueryCount(),

						// Networker
						b->getNWBytesIn() / (1024 * 1024),
						b->getNWBytesOut() / (1024 * 1024)
						);
			break;
		default		:
				sprintf(outbuf, "results, hosts, connections, uploads, "
					"downloads, options\r\nqueries, stats, shares "
					"or internals expected.\r\n");
				return 0;
	}

	return 1;
}

int Commander::doDownload(char *buf, Backend *b, char *outbuf)
{
	int	p;

	p	= buf[0];
	while ((*buf != ' ') && (*buf))
		buf++;
	while ((*buf == ' ') && (*buf))
		buf++;

	switch (p)
	{
		case	's'	:
			if (!b->startDownload(atoi(buf)))
			{
				sprintf(outbuf, "Start download of #%d FAILED!!!\r\n", atoi(buf));
				return 0;
			}
			break;
		case	'c'	:
			if (!b->continueDownload(atoi(buf)))
			{
				sprintf(outbuf, "Continue download of #%d FAILED!!!\r\n", atoi(buf));
				return 0;
			}
			break;
		case	'p'	:
			if (!b->pushDownload(atoi(buf)))
			{
				sprintf(outbuf, "Push download of #%d FAILED!!!\r\n", atoi(buf));
				return 0;
			}
			break;
		default	:
			sprintf(outbuf, "start, continue or push expected.\r\n");
			return 0;
	}

	return 1;
}

int Commander::doKill(char *buf, Backend *b, char *outbuf)
{
	int	p;

	p	= buf[0];
	while ((*buf != ' ') && (*buf))
		buf++;
	while ((*buf == ' ') && (*buf))
		buf++;

	switch (p)
	{
		case	'd'	:
			if (b->killDownload(atoi(buf)) == -1)
				sprintf(outbuf, "Download #%d not found!!!\r\n",
						atoi(buf));
			break;
		case	'u'	:
			if (b->killUpload(atoi(buf)) == -1)
				sprintf(outbuf, "Upload #%d not found!!!\r\n",
						atoi(buf));
			break;
		case	'c'	:
			if (!b->closeConnection(atoi(buf)))
				MCC->addError("connection not found.");
			break;
		default		:
			MCC->addError("\"download\", \"upload\" "
					"or \"connection\" expected.");
			return 0;
	}

	return 1;
}

int Commander::doShare(char *buf, Backend *b, char *outbuf)
{
	char	*c;
	int	p;

	p	= buf[0];
	while ((*buf != ' ') && (*buf))
		buf++;
	while ((*buf == ' ') && (*buf))
		buf++;

	if ((c = strchr(buf, '\r')))
		*c	= 0;
	if ((c = strchr(buf, '\n')))
		*c	= 0;

	switch (p)
	{
		case	'a'	:	b->addPath(buf);
			break;
		case	'd'	:	b->delPath(buf);
			break;
		case	'c'	:	b->delAllPaths();
			break;
		default		:
				MCC->addError("\"add\", \"del\" or \"clear\" "
						"expected.");
			return 0;
	}
	
	return 1;
}

int Commander::doScan(char *buf, Backend *b, char *outbuf)
{
	b->scanDirs();
	return 1;
}

int Commander::doSave(char *buf, Backend *b, char *outbuf)
{
	char	fn[FILE_NAME_LENGTH];
	int	p;

	p	= buf[0];
	while ((*buf != ' ') && (*buf))
		buf++;
	while ((*buf == ' ') && (*buf))
		buf++;

	switch (p)
	{
		case 'h'	:
			sprintf(fn, "%s/.hagelslag/hagelslag.hosts",
							getenv("HOME"));
			if (b->writeHosts(fn) == -1)
			{
				MCC->addError("couldn't save hosts!");
				return 0;
			}
			break;
		case 'o'	:
			sprintf(fn, "%s/.hagelslag/hagelslag.conf",
							getenv("HOME"));
			if (!b->writeOptions(fn))	
			{
				MCC->addError("couldn't save options!");
				return 0;
			}
			break;
		case 's'	:
			sprintf(fn, "%s/.hagelslag/hagelslag.shares",
							getenv("HOME"));
			b->writeShares(fn);
			break;
		default		:
				MCC->addError("\"hosts\",\"options\" or "
						"\"shares\" expected.");
			return 0;
	}

	return 1;
}

void Commander::getStateName(int st, char *buf)
{
        switch (st)
        {
                case    CON_STATE_INACTIVE      :
                                sprintf(buf, "INACTIVE");
                                break;
                case    CON_STATE_INIT  :
                                sprintf(buf, "CONNECTING");
                                break;
                case    CON_STATE_CONNECTED     :
                                sprintf(buf, "CONNECTED");
				break;
                case    CON_STATE_LOGIN_SENT      :
                                sprintf(buf, "LOGIN_SENT");
                                break;
                case    CON_STATE_ACTIVE        :
                                sprintf(buf, "ACTIVE\t");
                                break;
                case    CON_STATE_DONE  :
                                sprintf(buf, "DONE\t");
                                break;
                case    CON_STATE_CON_FAILED    :
                                sprintf(buf, "CONNECTION FAILED");
                                break;
                case    CON_STATE_LOGIN_FAILED  :
                                sprintf(buf, "LOGIN FAILED");
                                break;
                case    CON_STATE_CON_ERROR     :
                                sprintf(buf, "CONNECTION ERROR");
                                break;
                case    CON_STATE_CANCEL        :
                                sprintf(buf, "CANCELLED");
                                break;
                case    CON_STATE_CLOSED        :
                                sprintf(buf, "CLOSED\t");
                                break;
                case    CON_STATE_PENDING      :
                                sprintf(buf, "PENDING");
                                break;
/*
                case    CON_STATE_      :
                                sprintf(buf, "");
                                break;
*/
                default : sprintf(buf, "INVALID");
        }
}

int Commander::interpCommand(char *buf)
{
	while (*buf == ' ')
		buf++;

	switch(buf[0])
	{
		case 'h'	:
			switch (buf[1])
			{
				case 'e'	: return COMMAND_HELP;
				case ' '	: return COMMAND_HOST;
				case 'o'	: if ( (buf[2] == 's') || (buf[2] == ' ') )
							return COMMAND_HOST;
						  if (buf[2] == 'l')
							return COMMAND_HOLDON;
				default		: return COMMAND_INVALID;
			}
		case 's'	:
			switch (buf[1])
			{
				case ' '	: return COMMAND_SEARCH;
				case 'e'	: if (buf[2] == 't')
							return COMMAND_SET;
						  if ( (buf[2] == 'a') || (buf[2] == ' ') )
							return COMMAND_SEARCH;
						  return COMMAND_INVALID;
				case 'c'	: return COMMAND_SCAN;
				case 'a'	: return COMMAND_SAVE;
				case 'h'	: return COMMAND_SHARE;
				default		: return COMMAND_INVALID;
			}
		case 'd'	:
			return COMMAND_DOWNLOAD;
		case 'l'	:
			return COMMAND_LIST;
		case 'k'	:
			return COMMAND_KILL;
		case 'v'	:
			return COMMAND_VERSION;
		case 'q'	:
			return COMMAND_QUIT;
		case 'c'	:		// scroll mode
			return COMMAND_LIST;
/*
		case ''	:
			return COMMAND_;
*/
		default	: break;
	}
	return COMMAND_INVALID;
}

void Commander::mkPrintable(char *buf)
{
	while (*buf)
	{
		if (	!((*buf >= 'A')	&& (*buf <= 'Z'))	&&
			!((*buf >= 'a')	&& (*buf <= 'z'))	&&
			!((*buf >= '0') && (*buf <= '9')) 	&&
			!strchr("_=~`|<>+- \\/()[]{};':!*.,@#$%^&\"", *buf) )
			*buf	='?';
		buf++;
	}
}
