/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "Constants.h"
#include "DBElement.h"
#include "Socket.h"
#include "Buffer.h"
#include "Connection.h"
#include "Transfer.h"
#include "Download.h"

Download::Download(int newFileID, char *newFileName,
			char *newDirName, int newOffset, char *newCID)
	: Transfer(newFileID, newFileName, newDirName, newOffset)
{
	type		= CON_TYPE_DOWNLOAD;
	memcpy(cid, newCID, CLIENTID_LENGTH);
	waitSecs	= 0;
}

Download::Download(Connection *c, int newFileID, char *newFileName,
			char *newDirName, int newOffset, char *newCID)
	: Transfer(newFileID, newFileName, newDirName, newOffset, c)
{
	type		= CON_TYPE_DOWNLOAD;
	memcpy(cid, newCID, CLIENTID_LENGTH);
	waitSecs	= 0;
}

Download::Download(Connection *c)
	: Transfer(-1, "", "", 0, c)
{
	type		= CON_TYPE_DOWNLOAD;
	waitSecs	= 0;
}

void Download::reInit(int newFileID, char *newFileName, char *newDirName,
				char *newCID)
{
	type		= CON_TYPE_DOWNLOAD;
	waitSecs	= 0;
	fileID		= newFileID;

	strncpy(fileName, newFileName, TRANSFER_FN_LENGTH - 2);
	fileName[TRANSFER_FN_LENGTH - 1]	= 0;
	strncpy(dirName, newDirName, TRANSFER_FN_LENGTH - 2);
	dirName[TRANSFER_FN_LENGTH - 1]		= 0;
	memcpy(cid, newCID, CLIENTID_LENGTH);
}
										
int Download::readBlob()
{
	int	   	bytesRead;
	int		i;

	switch(bytesRead	= this->read())
	{
		case	0 :	return 0;
		case	-1:
				if (bytesLeft <= 0)
					setState(CON_STATE_DONE);
				else
				{
					if (type == CON_TYPE_PUSH_IN)
						setState(CON_STATE_DONE);
					else
						setState(CON_STATE_CON_ERROR);
				}

				/*
				 * at most 5 attempts to write the buffer to file
				 */
				for(i = 0; i < 5; i++)
					if (inBuffer->bytesInBuffer() > 0)
						inBuffer->bufFWrite(file);
					else
						break;

    	            	return 0;
    	          default	:
				bytesLeft	-= bytesRead;
				bytesDone	+= bytesRead;
				inBuffer->bufFWrite(file);
	}
	
	return bytesRead;
}

int Download::waitResponse()
{
	char	getMessage[512];
	int	res;
	char	*str;

	if (getInBufferBytes() <= 0)
		return 0;

	if (((res = peekInBuffer(getMessage, 511))) < 34) // 34 is an estimated minimum size for a http header
		return 0;

	getMessage[res]	= 0;
	if (!((str = strstr(getMessage, "\r\n\r\n"))))
		return 0;

	if (!strstr(getMessage, "HTTP") ||  !strstr(getMessage, "200"))
	{
		if (	strstr(getMessage, "503")										||
				(strstr(getMessage, "404") && strstr(getMessage, "maximum"))	)
		{
				setState(CON_STATE_PENDING);
				setWaitSecs(time(0));
				close();
				return 0;
		}
		setState(CON_STATE_LOGIN_FAILED);
 		return -1;
	}
	if ((str = strstr(getMessage, "Content-range: bytes=")))
	{
		int	start;
		int	end;

		str 		+= strlen("Content-range: bytes=");
		sscanf(str, "%d-%d/%d", &start, &end, &bytesTotal);
		bytesDone	= 0;
	}
	else if ((str = strstr(getMessage, "Content-length:")))
	{
		str 		+= strlen("Content-length:");
		bytesLeft	= bytesTotal	= atoi(str);
		bytesDone	= 0;
	}
	if (!((str = strstr(getMessage, "\r\n\r\n"))))
	{
		fprintf(stderr, "Error : BUG : lost end of HTTP HEADER during parse in "
				"Download::waitResponse()!!!\r\n");
		setState(CON_STATE_LOGIN_FAILED);
 		return 0;
	}
	str	+= strlen("\r\n\r\n");

	/*
	 * Remove the HTTP header
	 */
	if (res < (int)(str - getMessage))
	{
		fprintf(stderr, "Error : BUG : Download : %d > %d!\n",
				(int)(str - getMessage), res);
		setState(CON_STATE_LOGIN_FAILED);
		return -2;
	}

	if (skipBytes((int)(str - getMessage)) != (int)(str - getMessage))
	{
		setState(CON_STATE_LOGIN_FAILED);
		fprintf(stderr, "Error : BUG : Download : Remove HTTP HEADER failed !!\n");
		return -3;
	}

	setState(CON_STATE_ACTIVE);
	return 1;
}

int Download::loginToPeer()
{
	char	getMessage[512];

	if (offset <= 0)
		sprintf(getMessage, 	"GET /get/%d/%s HTTP/1.0\r\n"
					"Connection: Keep-Alive\r\n"
					"User-Agent: Hagelslag\r\n\r\n",
					fileID, fileName);
	else
		sprintf(getMessage, 	"GET /get/%d/%s HTTP/1.0\r\n"
					"Connection: Keep-Alive\r\n"
					"User-Agent: Hagelslag\r\n"
					"Range: bytes=%u-\r\n\r\n",
					fileID, fileName, offset);

	if (!sendStr(getMessage, strlen(getMessage)))
	{
		setState(CON_STATE_LOGIN_FAILED);
//		fprintf(stderr, "Error : Download : send Request failed !!!\n");
		return 0;
	}

	setState(CON_STATE_LOGIN_SENT);
	return 1;
}

void Download::getCID(char *buf)
{
	memcpy(buf, cid, CLIENTID_LENGTH);
}

time_t Download::getWaitSecs()
{
	return waitSecs;
}

int Download::getResultConnection()
{
	return resultConnection;
}

void Download::setWaitSecs(time_t t)
{
	waitSecs	= t;
}

void Download::setResultConnection(int i)
{
	resultConnection	= i;
}

/*
 * FAME
 */

char *Download::getCIDp()
{
	return (char *)cid;
}
