/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <time.h>

#include "Constants.h"
#include "ControlCentre.h"
#include "DBElement.h"
#include "DB.h"
#include "Socket.h"
#include "Buffer.h"
#include "Connection.h"
#include "FSEntry.h"
#include "Path.h"
#include "File.h"
#include "FSScanner.h"

int FSScanner::scanDir(Path *paths, DB *files)
{
	Path	*p;
	Path	*prv;
	Path	*last;
	
        if (!paths)
        	return files->getElements();

        /*
         * Remove all subdirs to speed things up. They'll be added
         * in doScanRec anyway
         */
        p       = last	= prv   = paths;
        while (p)
        {
                if (p->isSubdir())
                {
                        prv->setNext(p->getNext());
                        if (p == prv)
                        {
                                p       = (Path *)prv->getNext();
                                delete prv;
                                paths   = prv   = p;
                        }
                        else
                        {
                                delete p;
                                p       = (Path *)prv->getNext();
                        }
                }
                else
                {
                        prv     = p;
                        p       = (Path *)p->getNext();
                }
                if (p)
                	last	= p;
        }
        
        if (!paths)
        	return files->getElements();

        scanDirIt(paths, files, last);
        
        /*
         * Call an expensive balance to optimize the DB for future use.
         * The DB won't change often, so doing this once for every scan should
         * improve performance.
         */
 
	files->balance();                
 
        return files->getElements();
}

/*
 * Iterative breadth-first scan.
 */
int FSScanner::scanDirIt(Path *paths, DB *files, Path *last)
{
	Path	*subdirs;
	Path	*p;
	int	res;

	subdirs	= 0;
	res	= 0;

	for (p = paths; p; p = (Path *)p->getNext())
	{
		FSEntry		*fse;
		FSEntry		*tmp;

		fse	= p->list();
	
		while (fse)
		{
			if (fse->getMagic() == FS_DIRECTORY)
			{
				last->setNext((Path *)fse);
				last	= (Path *)fse;
				fse = fse->getNext();
				continue;
			}
			
			if (fse->getMagic() == FS_FILE)
			{
				if (((File *)fse)->getSize() == FILE_SIZE_UNKNOWN)
				{
					tmp	= fse->getNext();	
					delete fse;
					fse	= tmp;
					continue;
				}

				if (files->storeElement((File *)fse) == ENTRY_INVALID)
				{
/*					
					fprintf(stderr, "ERROR : BUG : FSScanner::scanDirRec() : "
							"Can't store file!!!\r\n");
*/					
					tmp	= fse->getNext();	
					delete fse;
					fse	= tmp;
					continue;
				}
			
				res++;
				
				if (files->isFull())
				{
					while (fse)
					{
						tmp	= fse;
						fse	= fse->getNext();
						delete tmp;
					}
					return res;
				}
				
				fse = fse->getNext();
				continue;
			}
			
			fprintf(stderr, "ERROR : BUG : FSScanner::scanDirRec() : fse->magic = %d\r\n!!!\n",
												fse->getMagic());
			tmp	= fse;
			fse	= fse->getNext();	
			delete tmp;
			continue;			
		}
	}
	
	return res;
}
