/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "Constants.h"
#include "ControlCentre.h"
#include "DBElement.h"
#include "DB.h"
#include "Socket.h"
#include "Buffer.h"
#include "GnutellaPacket.h"
#include "Connection.h"
#include "Packet.h"
#include "Host.h"
#include "FSEntry.h"
#include "Path.h"
#include "File.h"
#include "CacheEntry.h"
#include "Transfer.h"
#include "Upload.h"
#include "TransferStatus.h"
#include "LocalResult.h"
#include "FSScanner.h"
#include "HTTPServer.h"
#include "FServer.h"

#define TOUPPER(X)	( ( ((X) >= 'a') && ((X) <= 'z') ) ? \
				((X) + ('A'-'a')) : (X) )

#define BIT_ARRAY_LENGTH	((files->getElements() / UIBITS) + 1)

#define MKARRAY(A,X,T)	\
		{if (!A[X])						\
		{							\
			int	sz;					\
									\
			sz	= BIT_ARRAY_LENGTH;	\
 			A[X]	= (unsigned int *)			\
 					calloc(sz, sizeof(unsigned int)); \
			T	+= sz;					\
			if (!A[X])					\
			{						\
				fprintf(stderr, "Calloc failed!!!");	\
				return 0;				\
			}						\
  		}}
  		
#define CRASH()		{*(int*)0=0;}

FServer::FServer(ControlCentre *newMCC)
{
	int	i;

	MCC		= newMCC;
	uploads		= new DB(MCC->get_MAX_UPLOADS(), HASH_TINY);
	if (MCC->get_CACHE_ENTRIES() > 0)
		cache		= new DB(MCC->get_CACHE_ENTRIES(), HASH_LARGE);
	else
		cache	= 0;
	files		= 0;
	sizeShared	= 0;
	paths		= 0;
	localSearches	= 0;
	cacheHits	= 0;
	searchArraySize	= 0;
	hashArraySize	= 0;
	strcpy(vc, VALIDCHARS);

	BITS[0]		= 1;
	for (i = 1; i < (int)UIBITS; ++i)
		BITS[i]	= 2 * BITS[i - 1];

	for (i = 0; i < (int)SEARCH_ARRAY_SIZE; ++i)
	{
		searchArray[i]	= 0;
		searchCRC[i]	= 0;
		searchCount[i]	= 0;
	}

	for (i = 0; i < (int)HASH_ARRAY_SIZE; ++i)
	{
		hashArray[i]	= 0;
		hashCRC[i]	= 0;
		hashCount[i]	= 0;
	}
}

Path *FServer::getPaths()
{
	return paths;
}

int FServer::addUpload(Connection *c)
{
	Upload	*u;
	
	if (uploads->isFull())
	{
		SEND503(c);
		delete c;
		return 0;
	}

	if ((u = this->serve(c, files)))
		if (uploads->storeElement(u) == ENTRY_INVALID)
		{
			/*
			 * This is theoretically impossible.
			 */
			c->close();
			delete c;
			delete u ;
			return 0;
		}

	u->setCID(MCC->get_CLIENTIDp());
	
	return 1;		
}

int FServer::doScan()
{
	File	*f;
	int	i;
	int	elems;

	/*
	 * Clean up
	 */
	if (files)
	{
		files->clearDB();

		// Sanity check...
		if (	(MCC->get_MAX_FILES_SHARED() != files->getSize()) ||
								(!paths)   )
		{
			MCC->addError("FServer::doScan() : MAX_FILES_SHARED "
							"!= files->getSize()");
			delete files;
			files	= 0;
		}
	}

	if (cache)
		cache->clearDB();

	for (i = 0; i < (int)SEARCH_ARRAY_SIZE; ++i)
		if (searchArray[i])
		{
			free(searchArray[i]);
			searchArray[i]	= 0;
		}

	for (i = 0; i < (int)HASH_ARRAY_SIZE; ++i)
		if (hashArray[i])
		{
			free(hashArray[i]);
			hashArray[i]	= 0;
		}

	sizeShared	= 0;
	searchArraySize	= 0;
	hashArraySize	= 0;

        /*
         * Rebuild
         */
        if ((MCC->get_MAX_FILES_SHARED() <= 0) || (!paths))
        	return 0;

	if (!files)
		files	= new DB(HASH_MEDIUM, MCC->get_MAX_FILES_SHARED());
                 
   
	if ((elems = scanDir(paths, files)) > 0)
        {
		for(i = 0; i < elems; ++i)
                	if ((f = (File *)files->findElement(i)))
                        {
	               		addToHashArray(f->getPreparedFnp(), f->getEntry());
	               		addToSearchArray(f->getPreparedFnp(), f->getEntry());
				sizeShared	+= f->getSize();
        		}
        }
        else
        {
        	files->clearDB();
        	delete files;
        	files	= 0;
        }

	return 0;
}

int FServer::push(int fileID, unsigned int hostip,
					unsigned short hostport)
{
	Upload			*u;
	File			*f;
	Path			*p;
	
	/*
	 * See if we can find that file....
	 */

	if (!files)
		return 0;
		
	if (!(f = (File *)files->findElement(fileID)))
		return 0;

	if (!(p = f->getPath()))
		return 0;

	u	= new Upload(fileID, f->getFileNamep(), p->getPathp(), 0, MCC->get_CLIENTIDp());
	u->setType(CON_TYPE_PUSH_OUT);
	u->setOutgoing();
	
	if ( (!u->Start(LOCAL_READ)) || (uploads->storeElement(u) == ENTRY_INVALID))
	{
		delete u;
		return 0;
	}

	u->Connect(hostip, hostport ? hostport : 6346);
	u->tryToConnect();
	return 1;
}

int FServer::pollFServer()
{
	return handleUploads();
}

int FServer::handleUploads()
{
	int	i;
	int	t;
	Upload	*u;

	if (!uploads->getElements())
		return 0;

	t	= 0;
	for (i = 0; i < uploads->getSize(); ++i)
	{
		if ( (u = (Upload *)uploads->findElement(i)) )
		{
			switch(u->getState())
			{
				case	CON_STATE_INIT		:
						u->tryToConnect();
						break;
				case	CON_STATE_CONNECTED	:
						u->loginToPeer();
						break;
				case	CON_STATE_LOGIN_SENT	:
				case	CON_STATE_ACTIVE 	:
						if (u->sendBlob() >= 0)
							t++;
						break;
				case	CON_STATE_DONE		:
						/*
						 * Stupid Workaround...
						 * Without sleep the last data
						 * is not received correctly
						 */
//						sleep(1);
						uploads->deleteElement(i);
						break;
				default			:
						uploads->deleteElement(i);
					break;
			}
		}
	}

	return t;
}

int FServer::killUpload(int i)
{
	return uploads->deleteElement(i);
}

int FServer::addPath(char *newPath)
{
	Path	*np;
	Path	*check;
	char	pn[FILE_NAME_LENGTH];

	if (newPath)
	{
		if (newPath[0] == '~')
		{
#warning FServer::addPath : OS dependent code
			sprintf(pn, "%s%s", getenv("HOME"), newPath + 1);
			newPath		= &pn[0];
		}
		
		for (check = paths; check; check = (Path *)check->getNext())
		{
			if (!strncmp(check->getPathp(), newPath, PATH_NAME_LENGTH))
			{
				check->setSubdir(0);
				return 1;
			}
		}
		
		np	= new Path(newPath);
		np->setNext(paths);
		paths	= np;
		return 1;
	}

	return 0;
}

int FServer::deletePath(char *oldPath)
{
	Path	*prv, *tmp;
	int	res;

	prv	= 0;
	res	= 0;
	for (tmp = paths; tmp; tmp = (Path *)tmp->getNext())
	{
		if (strlen(tmp->getPathp()) < strlen(oldPath))
		{
			prv	= tmp;
			continue;
		}
		if (!strncmp(tmp->getPathp(), oldPath, strlen(oldPath)))
		{
			res	= 1;
			if (prv)
				prv->setNext(tmp->getNext());
			else
				paths	= (Path *)tmp->getNext();
			delete tmp;
			if (prv)
				tmp	= prv;
			else
				tmp	= paths;

			if (files)
			{
				int	i;

				files->clearDB();
				delete(files);
				files	= 0;
				if (cache)
					cache->clearDB();

				for (i = 0; i < (int)SEARCH_ARRAY_SIZE; ++i)
					if (searchArray[i])
					{
						free(searchArray[i]);
						searchArray[i]	= 0;
					}
			}
		}

		if (tmp)
			prv	= tmp;
		else
			break;
	}
	
	return res;
}

int FServer::delAllPath()
{
	Path	*p, *prv;

	p	= paths;
	while (p)
	{
		prv	= p;
		p	= (Path *)p->getNext();
		delete	prv;
	}
	paths	= 0;

	return 1;
}

TransferStatus *FServer::getStatusList()
{
	Upload		*u;
	TransferStatus	*prv, *current;
	int		i;

	current		= 0;
	for (i = 0; i < uploads->getSize(); i++)
		if ((u = (Upload *)uploads->findElement(i)))
		{
			prv	= current;
			current	= new TransferStatus(u);
			current->setNext(prv);
		}

	return current;
}

unsigned int FServer::getSizeShared()
{
	return sizeShared;
}

unsigned int FServer::getFilesCount()
{
	if (files)
		return (unsigned int)files->getElements();

	return 0;
}

unsigned int FServer::getLocalSearches()
{
	return localSearches;
}

unsigned int FServer::getCacheHits()
{
	return cacheHits;
}

unsigned int FServer::getCacheEntries()
{
	return cache->getElements();
}

unsigned int FServer::getCacheSize()
{
	return cache->getSize();
}

unsigned int FServer::getSearchArraySize()
{
	return searchArraySize;
}

unsigned int FServer::getHashArraySize()
{
	return hashArraySize;
}

int FServer::resize()
{
        DB      *newUploads;
        int     i;

        if (MCC->get_MAX_UPLOADS() < uploads->getElements())
	{
                MCC->addError("FServer : Too many uploads for "
                                "resize!");
                return 0;
        }

        newUploads    = new DB(MCC->get_MAX_UPLOADS(), HASH_TINY);

        for (i = 0; (	(i < newUploads->getSize())	&&
			(i < uploads->getSize())	); i++)
                newUploads->storeElement(uploads->findElement(i));

        delete uploads;

        uploads       = newUploads;

        return 1;
}

int FServer::resizeCache()
{
	int	i;
	DB 	*newCache;

	if (MCC->get_CACHE_ENTRIES() <= 0)
	{
		if (!cache)
			return 0;
		cache->clearDB();
		delete cache;
		cache	= 0;
		return 0;
	}

	if (cache->getSize() == MCC->get_CACHE_ENTRIES())
		return 0;

	newCache	= new DB(MCC->get_CACHE_ENTRIES(), HASH_LARGE);

	for (i = 0; (i < cache->getSize()) && (i < newCache->getSize()); ++i )
	{
		CacheEntry	*ce;

		if ((ce	= (CacheEntry *)cache->findElement(i)))
		{
			cache->removeElement(i);
			if (newCache->storeElement(ce) == ENTRY_INVALID)
				delete ce;
		}
	}
	cache->clearDB();
	delete cache;
	cache	= newCache;
	return 1;
}

int FServer::isUsingHost(int hi)
{
	int		i;
	Upload		*u;

	for (i = 0; i < uploads->getSize(); i++)
	{
		if ((u = (Upload *)uploads->findElement(i)))
			if (u->getHostIndex() == hi)
				return 1;
	}

	return 0;
}

/*
 * This thing used practically all CPU power. Time for optimization!!!
 */

Packet *FServer::mkSearchResp(char *findme)
{
	Packet		*p;
	LocalResult	*res, *wres;
	int		i;
	int		more;
	char		fn[FILE_NAME_LENGTH];
	CacheEntry	*ce;
	unsigned int	*searchBits;

	if (!files)
		return 0;

	localSearches++;
	more	= preparePattern(findme);

	if ( (!more) || (strlen(findme) < 3) )
		return 0;

	/*
	 * check cache
	 */
	if (cache)
	{
		ce	= new CacheEntry(findme, more);

		if (((i = cache->containsElement(ce))) != ENTRY_INVALID)
		{
			delete ce;
			cacheHits++;
			cache->touch(i);
			if ((ce = (CacheEntry *)cache->findElement(i)))
				return ce->getResponsep();
			
			MCC->addError("BUG : FServer : cacheEntry expected "
						"but not found");

			return 0;
		}

		if (cache->storeElement(ce) == ENTRY_INVALID)
		{
			delete ce;
			ce	= 0;
			MCC->addError("BUG : Can't store CacheEntry");
			return 0;
		}
	}
	else
	{
		/*
		 * BUGFIX : Continuing could lead to a mem leak as
		 * the resulting packet would never be deleted.
		 */
		return 0;
	}
	
	/*
	 * start search
	 */
	searchBits	= searchHashArray(findme, 0);
	if (!((searchBits = searchSearchArray(findme, searchBits))))
		return 0;

	/*
	 * Create LocalResult List
	 */
	res	= mkLocalResultList(searchBits, findme);
	free(searchBits);

	if (!res)
		return 0;

	/*
	 * Create packet.
	 */
	p	= new Packet(MCC->get_DEFAULT_TTL());
	p->mkSearchRespPacket(MCC->get_CLIENTIDp(), MCC->get_PORT(),
				0, MCC->get_SPEED(), MCC->get_DEFAULT_TTL());

	for (wres = res; wres; wres = wres->getNext())
	{
		(wres->getLRFile())->getFileName(fn);
		if (!p->addToSRPacket((wres->getLRFile())->getEntry(),
					(wres->getLRFile())->getSize(),
					fn))
 			break;
	}

	/*
	 * Attach packet to Cache Entry
	 */
	if (ce)
		ce->setResponse(p);

	/*
	 * Cleanup
	 */
	res->recDelLocalResult();

	return p;
}

unsigned int *FServer::ANDSearch(unsigned int *result, unsigned int *model,
					unsigned int range)
{
	unsigned int	i;
	int		res;

	if (!model)
	{
		if (result)
			free(result);
		return 0;
	}

	if (!result)
	{
		result	= (unsigned int *)malloc(range * sizeof(unsigned int));
		memcpy(result, model, range * sizeof(unsigned int));
		return result;
	}

	res	= 0;
	for (i = 0; i < range; ++i)
		res = ( res |	((result[i]	= result[i] & model[i])) );

	if (!res)
	{
		free(result);
		return 0;
	}

	return result;
}


LocalResult *FServer::mkLocalResultList(unsigned int *searchBits, char *findme)
{
	LocalResult	*res;
	LocalResult	*wres;
	LocalResult	*prv;
	File		*f;
	char		*next;
	unsigned int	i;
	int		more;

	res	= 0;
	more	= 1;

	/*
	 * Isolate first pattern, next is beginning of the rest
	 */
	for (next = findme; 1; next++)
	{
		if (*next)
		{
			if (strchr(SEPARATORS, (int)*next))
			{
				*next	= 0;
				break;
			}
		}
		else
		{
			more	= 0;
			break;
		}
	}
	next++;

	for (i = 0; i < (unsigned int)files->getElements(); i++)
	{
		// check to see if any bits in int are set
		if (!searchBits[ i / UIBITS ])
		{
			i	+= UIBITS - 1;
			continue;
		}

		// if so, find a bit
		if (!ISSET(searchBits, i))
			continue;

		// find the corresponding file
		if (!((f = (File *)files->findElement(i))))
		{
 			MCC->addError("BUG : FServer::mkLocalResultList() : Inconsistency in FileServer");
 			continue;
		}

		// And add it to our localresult list.
		if (strstr(f->getPreparedFnp(), findme))
			res	= new LocalResult(f, res);
	}

	if (!res)
		return 0;

	/*
	 * Scan list for the rest of the patterns
	 */
	prv	= 0;
	while(more)
	{
		findme	= next;

		/*
		 * Find end of pattern & next
		 */
		while(1)
		{
			if (*next)
			{
				if (strchr(SEPARATORS, (int)*next))
				{
					*next	= 0;
					break;
				}
				*next	= TOUPPER(*next);
			}
			else
			{
				more	= 0;
				break;
			}
			next++;
		}
		next++;

		wres	= res;
		prv	= 0;
		while (wres)
		{
			if (strstr(wres->getFileNamep(), findme))
			{
				prv	= wres;
				wres	= wres->getNext();
			}
			else
			{
				/*
				 * Pattern not found, delete wres from list
				 */
				if (!prv)
				{
					prv	= wres;
					wres	= wres->getNext();
					res	= wres;
					delete	prv;
					prv	= 0;
				}
				else
				{
					wres	= wres->getNext();
					delete prv->getNext();
					prv->setNext(wres);
				}
			}
		}

		if (!res)
			more	= 0;
	}
	return res;
}

unsigned int *FServer::searchHashArray(char *findme,
					unsigned int *searchBits)
{
	unsigned char	*c;
	int		i;
	int		hv;
	unsigned int	beenThereDoneThat[1 + (HASH_ARRAY_SIZE / UIBITS)];

	if (strlen(findme) < 4)
		return searchBits;

 	for (i = 0; i < (int)(1 + (HASH_ARRAY_SIZE / UIBITS)); ++i)
		beenThereDoneThat[i]	= 0;

	while ((c = (unsigned char *)strchr(findme, ' ')))
	{
		int	l;

		*c	= 0;

		if ((l = strlen(findme)) > 3)
			hv	= hashStr(findme, HASH_ARRAY_SIZE);
		else
		{
			findme	+= l + 1;
			*c	= ' ';
			continue;
		}

		if ( (hashArray[hv]) && (!ISSET(beenThereDoneThat, hv)) )
		{
			searchBits = ANDSearch(searchBits, hashArray[hv], BIT_ARRAY_LENGTH);
			
			if (!searchBits)
			{
				*c	= ' ';
				return 0;
			}

			SETBIT(beenThereDoneThat, hv);
		}
		findme	+= l + 1;
		*c	= ' ';
	}
	if (strlen(findme) > 3)
	{
		hv	= hashStr(findme, HASH_ARRAY_SIZE);
		
		if ( (!ISSET(beenThereDoneThat, hv)) && (hashArray[hv]) )
			searchBits = ANDSearch(searchBits, hashArray[hv], BIT_ARRAY_LENGTH);
	}

	return searchBits;
}

unsigned int *FServer::searchSearchArray(char *findme,
					unsigned int *searchBits)
{
	int		i;
	int		inc;
	unsigned int	beenThereDoneThat[1 + (SEARCH_ARRAY_SIZE / UIBITS)];

 	for (i = 0; i < (int)(1 + (SEARCH_ARRAY_SIZE / UIBITS)); ++i)
		beenThereDoneThat[i]	= 0;

	for (i = 0; i < (int)strlen(findme); i += inc)
	{
		int sum;

		inc	= 1;
		if  ( (findme[i]) && (findme[i] != ' ') )
		{
			if ( (findme[i + 1])  && (findme[i + 1] != ' ') )
			{
				if ( (findme[i + 2]) && (findme[i + 2] != ' ') )
				{
					sum	= VAL3(findme[i], findme[i + 1], findme[i + 2]);
					inc	= 3;
				}
				else
				{
					sum	= VAL2(findme[i], findme[i + 1]);
					inc	= 2;
				}
			}
			else
				sum	= VAL1(findme[i]);

			if (ISSET(beenThereDoneThat, sum))
				continue;

			searchBits = ANDSearch(searchBits, searchArray[sum], BIT_ARRAY_LENGTH);
			
			if (!searchBits)
				return 0;

			SETBIT(beenThereDoneThat, sum);
		}
	}
	return searchBits;
}

int FServer::addToHashArray(char *pattern, int idx)
{
	int		res;
	int		hv;
	unsigned char	*c;

	if (strlen(pattern) < 4)
		return 0;

	res	= 0;

	while ((c = (unsigned char *)strchr(pattern, ' ')))
	{
		int	l;

		*c	= 0;
		if ((l = strlen(pattern)) > 3)
		{
			res	= 1;
			hv	= hashStr(pattern, HASH_ARRAY_SIZE);
			if ( (hv < 0) || (hv >= HASH_ARRAY_SIZE) )
				CRASH();
			MKARRAY(hashArray, hv, hashArraySize);
//			if (!ISSET(hashArray[hv], idx))
//				hashCount[hv]++;
			SETBIT(hashArray[hv], idx);
//			hashCRC[hv]	= crc(hashArray[hv], BIT_ARRAY_LENGTH);
		}
		pattern	+= l + 1;
		*c	= ' ';
	}
	
	if (strlen(pattern) > 3)
	{
		res	= 1;
		hv	= hashStr(pattern, HASH_ARRAY_SIZE);
			if ( (hv < 0) || (hv >= HASH_ARRAY_SIZE) )
				CRASH();
		MKARRAY(hashArray, hv, hashArraySize);
//		if (!ISSET(hashArray[hv], idx))
//			hashCount[hv]++;
		SETBIT(hashArray[hv], idx);
//		hashCRC[hv]	= crc(hashArray[hv], BIT_ARRAY_LENGTH);
	}
	
	return res;
}

int FServer::addToSearchArray(char *fn, int idx)
{
	int	i;
	int	len;
	int 	sum;
	
	len	= strlen(fn);
	for (i = 0; i < len; ++i)
		if ( (fn[i]) && (fn[i] != ' ') )
		{
			sum	= VAL1(fn[i]);
			MKARRAY(searchArray, sum, searchArraySize);
//			if (!ISSET(searchArray[sum], idx))
//				searchCount[sum]++;
   			SETBIT(searchArray[sum], idx);
//			searchCRC[sum]	= crc(searchArray[sum], BIT_ARRAY_LENGTH);

 			if ( (fn[i + 1]) && (fn[i + 1] != ' ') )
			{
				sum	= VAL2(fn[i], fn[i + 1]);
				MKARRAY(searchArray, sum, searchArraySize);
//				if (!ISSET(searchArray[sum], idx))
//					searchCount[sum]++;
				SETBIT(searchArray[sum], idx);
//				searchCRC[sum]	= crc(searchArray[sum], BIT_ARRAY_LENGTH);

 				if ( (fn[i + 2]) && (fn[i + 2] != ' ') )
				{
					sum	= VAL3(fn[i], fn[i + 1], fn[i + 2]);
					MKARRAY(searchArray, sum, searchArraySize);
//					if (!ISSET(searchArray[sum], idx))
//						searchCount[sum]++;
					SETBIT(searchArray[sum], idx);
//					searchCRC[sum]	= crc(searchArray[sum], BIT_ARRAY_LENGTH);
				}
			}
		}

	return 1;
}

int FServer::preparePattern(char *buf)
{
	int	i;
	int	j;
	int	res;

	res	= 0;
	j	= 0;
	for (i = 0; buf[i] && (i < FILE_NAME_LENGTH); ++i)
	{
		buf[i + j]	= TOUPPER(buf[i]);

		if (strchr(VALIDCHARS, buf[i + j]))
		{
			if (	((i + j) > 0)	&& (buf[i + j - 1] == ' ') ||
				(!i)	)
				res++;
		}
		else
		{
			buf[i + j]	= ' ';
			if (	( (i > 0) && (buf[i + j - 1] == ' ') ) 	||
				( (i == 0) && (buf[i + j] == ' '))	)
				j--;
		}
	}
	if (buf[i + j - 1] == ' ')
		buf[i + j - 1]	= 0;
	else
		buf[i + j]	= 0;

	return res;
}

unsigned int FServer::hashStr(char *buf, unsigned int prime)
{
	unsigned int	r;

	for (r = 0; *buf; buf++)
		r	= ((r * (*buf)) + (*buf)) % prime;

	return r;
}

unsigned int FServer::crc(unsigned int *array, int range)
{
	int		i;
	unsigned int	sum;

	if ( (!array) || (range <= 0) )
		return 0;

	sum	= 1;
	for (i = 0; i < range; ++i)
		sum	+= array[i];
	
	return sum;
}

int FServer::checkCRC()
{
	int		i;
	int		j;
	int		ok;
	unsigned int	count;
	
	ok	= 1;
	for (i = 0; i < HASH_ARRAY_SIZE; ++i)
	{
		if (crc(hashArray[i], BIT_ARRAY_LENGTH) != hashCRC[i])
		{
			fprintf(stderr, "ERROR : BUG : FServer::checkCRC() : hashArray[%d] failed CRC!!!\r\n", i);
			ok	= 0;
			CRASH();
		}
		
		if (!hashArray[i])
		{
			if (hashCount[i])
			{
				fprintf(stderr, "ERROR : BUG : FServer::checkCRC() : hashArray[%d] Count non-zero!!!\r\n", i);
				ok	= 0;
				CRASH();
			}
			continue;
		}
			
		count	= 0;
		for (j = 0; j < files->getElements(); j++)
			if (ISSET(hashArray[i], j))
				count++;
				
		if (count != hashCount[i])
		{
			fprintf(stderr, "ERROR : BUG : FServer::checkCRC() : hashArray[%d] Count incorrect!!!\r\n", i);
			ok	= 0;
			CRASH();
		}
	}

	for (i = 0; i < SEARCH_ARRAY_SIZE; ++i)
	{
		if (crc(searchArray[i], BIT_ARRAY_LENGTH) != searchCRC[i])
		{			
			fprintf(stderr, "ERROR : BUG : FServer::checkCRC() : searchArray[%d] failed CRC!!!\r\n", i);
			ok	= 0;
			CRASH();
		}
		
		if (!searchArray[i])
		{
			if (searchCount[i])
			{
				fprintf(stderr, "ERROR : BUG : FServer::checkCRC() : searchArray[%d] Count non-zero!!!\r\n", i);
				ok	= 0;
				CRASH();
			}
			continue;
		}
			
		count	= 0;
		for (j = 0; j < files->getElements(); j++)
			if (ISSET(searchArray[i], j))
				count++;
				
		if (count != searchCount[i])
		{
			fprintf(stderr, "ERROR : BUG : FServer::checkCRC() : searchArray[%d] Count incorrect!!!\r\n", i);
			ok	= 0;
			CRASH();
		}
	}
	
	return ok;
}

int FServer::integrity()
{
	unsigned int	i;
	File		*f;
	unsigned int	*searchBits;
	int		ok;
	unsigned int	j;
	
	checkCRC();
	ok	= 1;
	if (files->getSize() > SEARCH_ARRAY_SIZE)
	{
		fprintf(stderr, "ERROR : BUG : FServer::integrity() : files "
				"DB size (%d) > SEARCH_ARRAY_SIZE (%d)!!!\r\n",
				files->getSize(), SEARCH_ARRAY_SIZE);
		CRASH();
		return 0;
	}
	
	if (files->getSize() > HASH_ARRAY_SIZE)
	{
		fprintf(stderr, "ERROR : BUG : FServer::integrity() : files DB "
				"size (%d) > HASH_ARRAY_SIZE (%d)!!!\r\n",
				files->getSize(), HASH_ARRAY_SIZE);
		CRASH();
		return 0;
	}
	
	if (!hashArray)
	{
		fprintf(stderr, "ERROR : BUG : FServer::integrity() : No hashArray!!!\r\n");
		CRASH();
		return 0;
	}
	
	if (!searchArray)
	{
		fprintf(stderr, "ERROR : BUG : FServer::integrity() : No hashArray!!!\r\n");
		CRASH();
		return 0;
	}
	
	checkCRC();
	for (i = 0; i < (unsigned int)files->getElements(); i++)
	{
		char		findme[257];
		LocalResult	*res;
		int		found;
		
		if (!((f = (File *)files->findElement(i))))
		{			
			for (j = 0; j < HASH_ARRAY_SIZE; j++)
			{
				if (	hashArray[j]		&&
					ISSET(hashArray[j], i)	)
				{
					fprintf(stderr, "ERROR : BUG : FServer::integrity() "
							": bit %d set in hashArray %d\r\n", i, j);
					ok	= 0;
					CRASH();
					continue;
				}
			}
			
			for (j = 0; j < SEARCH_ARRAY_SIZE; j++)
			{
				if (	searchArray[j]		&&
					ISSET(searchArray[j], i))
				{
					fprintf(stderr, "ERROR : BUG : FServer::integrity() "
							": bit %d set in searchArray %d\r\n", i, j);
					ok	= 0;
					CRASH();
					continue;
				}
			}
			continue;
		}
		else
		{
			fprintf(stderr, "%d ", i);
			f->getFileName(findme);
			fprintf(stderr, "(%s) ---> ", findme);
			preparePattern(findme);
			fprintf(stderr, "(%s)\r\n", findme);
		
			checkCRC();
			if (i != (unsigned int)f->getEntry())
			{
				fprintf(stderr, "ERROR : BUG : FServer::integrity() : index "
						"mismatch, found (%d) instead of (%d) for (%s)\r\n",
						f->getEntry(), i, findme);
				ok	= 0;
				CRASH();
				continue;
			}
			
			found	= 0;
			for (j = 0; j < HASH_ARRAY_SIZE; j++)
				if (	hashArray[j]		&&
					ISSET(hashArray[j], i)	)
				{
					found	= 1;
					break;
				}
			
			if (!found)
			{
				fprintf(stderr, "WARNING : FServer::integrity() : bit %d "
						"NOT set in hashArray for (%s)\r\n",
						i, findme);
				ok	= 0;
				continue;
			}
			
			found	= 0;
			for (j = 0; j < SEARCH_ARRAY_SIZE; j++)
				if (	searchArray[j]		&&
					ISSET(searchArray[j], i))
				{
					found	= 1;
					break;
				}
			
			if (!found)
			{
				fprintf(stderr, "ERROR : BUG : FServer::integrity() : bit %d "
						"NOT set in searchArray for (%s)\r\n",
						i, findme);
				ok	= 0;
				CRASH();
				continue;
			}
		}
			
		checkCRC();
		/*
		 * start search
		 */
		searchBits	= searchHashArray(findme, 0);
		checkCRC();

		if(!searchBits)
		{
			char	*c;
			char	*s;
			
			s	= findme;
			while ((c = strchr(findme, ' ')))
			{
				*c	= 0;
				if (strlen(s) > 3)
				{
					*c	= ' ';
					fprintf(stderr, "ERROR : BUG : FServer::integrity() : searchHashArray found nothing for (%s)\r\n", findme);
					ok	= 0;
					CRASH();
					continue;
				}
				s	= c + 1;
				*c	= ' ';
			}
			if (strlen(s) > 3)
			{
				*c	= ' ';
				fprintf(stderr, "ERROR : BUG : FServer::integrity() : searchHashArray found nothing for (%s)\r\n", findme);
				ok	= 0;
				CRASH();
				continue;
			}
			
			fprintf(stderr, "WARNING : FServer::integrity() : searchHashArray found nothing for (%s)\r\n", findme);
			ok	= 0;
			continue;
		}
		
		if (!(ISSET(searchBits, f->getEntry())))
		{
			fprintf(stderr, "ERROR : BUG : FServer::integrity() : searchHashArray failed "
					"for bit %d, (%s)\r\n", f->getEntry(), findme);
			ok	= 0;
			if (searchBits)
				free(searchBits);
			CRASH();
			continue;
		}

		checkCRC();
		searchBits	= searchSearchArray(findme, searchBits);
		checkCRC();
		if(!searchBits)
		{
			fprintf(stderr, "ERROR : BUG : FServer::integrity() : searchSearchArray found nothing for (%s)\r\n",  findme);
			ok	= 0;
			CRASH();
			continue;
		}
		
		if (!(ISSET(searchBits, f->getEntry())))
		{
			fprintf(stderr, "ERROR : BUG : FServer::integrity() : searchSearchArray failed for (%s)\r\n",  findme);
			ok	= 0;
			if (searchBits)
				free(searchBits);
			CRASH();
			continue;
		}
		
		if (!((res = mkLocalResultList(searchBits, findme))))
		{
			fprintf(stderr, "ERROR : BUG : FServer::integrity() : mkLocalResultList failed for (%s)\r\n",  findme);
			ok	= 0;
			if (searchBits)
				free(searchBits);
			CRASH();
			continue;
		}

		checkCRC();
		if (searchBits)
			free(searchBits);
		
		if (res)
			res->recDelLocalResult();
		checkCRC();
	}
	if (!checkCRC())
		ok	= 0;
		
	return ok;
}
