/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdio.h>

#if ((OS == LINUX) || (OS == FREEBSD))
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#endif

#include "Constants.h"
#include "DBElement.h"
#include "FSEntry.h"
#include "Path.h"
#include "File.h"

#define TOUPPER(X)	( ( ((X) >= 'a') && ((X) <= 'z') ) ? (X) + ('A' - 'a') : (X))

File::File(int s, char *fn, Path *newFPath) : FSEntry(FS_FILE)
{
	int	i, j;

	size	= s;
	strncpy(fileName, fn, FILE_NAME_LENGTH - 2);
	fileName[FILE_NAME_LENGTH - 1]	= 0;
	j	= 0;
	for (i = 0; fileName[i] && i < FILE_NAME_LENGTH; ++i)
	{
		preparedFn[i + j]	= TOUPPER(fileName[i]);
		if (!strchr(VALIDCHARS, preparedFn[i + j]))
		{
			preparedFn[i + j]	= ' ';
			if (	( (i > 0) && (preparedFn[i + j - 1] == ' ') ) 	||
				( (!i) 	&& (preparedFn[i + j] == ' '))		)
				j--;
		}
	}
	if (preparedFn[i + j - 1] == ' ')
		preparedFn[i + j - 1]	= 0;
	else
		preparedFn[i + j]	= 0;
	FPath				= newFPath;

	i	= strlen(fileName);
	while (i < 4)
		fileName[i++]	= 0;
}

void File::setSize(int s)
{
	size	= s;
}

int File::getSize()
{
	return size;
}

void File::getFileName(char *buf)
{
	strncpy(buf, fileName, FILE_NAME_LENGTH - 2);
	buf[FILE_NAME_LENGTH - 1]		= 0;
}

void File::setFileName(char *buf)
{
	int	i, j;

	strncpy(fileName, buf, FILE_NAME_LENGTH - 2);
	fileName[FILE_NAME_LENGTH - 1]	= 0;
	j	= 0;
	for (i = 0; fileName[i] && i < FILE_NAME_LENGTH; ++i)
	{
		preparedFn[i + j]	= TOUPPER(fileName[i]);
		if (!strchr(VALIDCHARS, preparedFn[i + j]))
		{
			preparedFn[i + j]	= ' ';
			if (	( (i > 0) && (preparedFn[i + j - 1] == ' ') ) 	||
				( (!i) 	&& (preparedFn[i] == ' '))		)
				j--;
		}
	}

	if (preparedFn[i + j - 1] == ' ')
		preparedFn[i + j - 1]	= 0;
	else
		preparedFn[i + j]	= 0;
}

void File::getPreparedFn(char *buf)
{
	strncpy(buf, preparedFn, FILE_NAME_LENGTH - 2);
	buf[FILE_NAME_LENGTH - 1]		= 0;
}

Path *File::getPath()
{
	return FPath;
}

void File::setPath(Path *newFPath)
{
	FPath	= newFPath;
}

#if ((OS == LINUX) || (OS == FREEBSD))
int File::statFile()
{
	char		name[512];
	struct stat	st;
	
	if (!FPath)
		return 0;

	sprintf(name, "%s/%s", FPath->getPathp(), fileName);
	if (stat(name, &st))
		return 0;
		
	size	= st.st_size;
	return 1;
}
#endif

/*
 * DBElement Inheritance
 */

int File::equals(DBElement *myFile)
{
	return (!strncmp(fileName, ((File *)myFile)->getFileNamep(),
			FILE_NAME_LENGTH));
}

int File::lessThan(DBElement *myFile)
{
	return (strncmp(fileName,
			((File *)myFile)->getFileNamep(), FILE_NAME_LENGTH)
							< 0);
}

int File::greaterThan(DBElement *myFile)
{
	return (strncmp(fileName,
				((File *)myFile)->getFileNamep(), FILE_NAME_LENGTH)
							> 0);
}

unsigned int File::hashValue()
{
	return (*(unsigned int *)fileName);
}

/*
 * FAME
 */

char *File::getFileNamep()
{
	return (char *)fileName;
}

char *File::getPreparedFnp()
{
	return (char *)preparedFn;
}

