/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include "Constants.h"
#include "DBElement.h"
#include "GnutellaPacket.h"

GnutellaPacket::GnutellaPacket(unsigned char maxTTL)
{
	newMesgID();
	header[HEADER_FUNCTION]	= FUNC_INVALID;
	GPinit(maxTTL);
	packetBody		= 0;
	sourceConnection	= ENTRY_INVALID;
	targetConnection	= ENTRY_INVALID;
}

GnutellaPacket::GnutellaPacket(int newCon, unsigned char maxTTL)
{
	newMesgID();
	header[HEADER_FUNCTION]	= FUNC_INVALID;
	GPinit(maxTTL);
	sourceConnection	= newCon;
	targetConnection	= ENTRY_INVALID;
}

GnutellaPacket::GnutellaPacket(GnutellaPacket *ask)
{
	int		func;

	ask->getHeader((unsigned char *)header);
	switch (ask->getType())
	{
		case FUNC_INIT		: func = FUNC_INIT_RESP; break;
		case FUNC_SEARCH	: func = FUNC_SEARCH_RESP; break;
//		case FUNC_	: func =; break;
		default		: func = FUNC_INVALID;
	}
	header[HEADER_FUNCTION]		= func;
	*(int *)&header[HEADER_DATALENGTH]	= 0;
	packetBody	= 0;
	targetConnection	= ask->getSourceConnection();
	sourceConnection	= ENTRY_INVALID;
}

GnutellaPacket::GnutellaPacket(int newPacketType, char *newBody, int len,
				unsigned char defaultTTL)
{
	newMesgID();
	GPinit(defaultTTL);
	switch (newPacketType)
	{
		case FUNC_INIT		:
				header[HEADER_FUNCTION] = FUNC_INIT;
				break;
		case FUNC_INIT_RESP	:
				header[HEADER_FUNCTION] = FUNC_INIT_RESP;
				break;
		case FUNC_PUSH		:
				header[HEADER_FUNCTION] = FUNC_PUSH;
				break;
		case FUNC_SEARCH	:
				header[HEADER_FUNCTION] = FUNC_SEARCH;
				break;
		case FUNC_SEARCH_RESP	:
				header[HEADER_FUNCTION] = FUNC_SEARCH_RESP;
				break;
		default 		: header[HEADER_FUNCTION] = FUNC_INVALID;
						break;
	}

	*(int *)&header[HEADER_DATALENGTH]	= len;
	if (newBody)
	{
		packetBody	= (unsigned char *)malloc(len);
		memcpy(packetBody, newBody, len);
	}
	else
	{
		packetBody			= 0;
		*(int *)&header[HEADER_DATALENGTH]	= 0;
	}
	sourceConnection	= ENTRY_INVALID;
	targetConnection	= ENTRY_INVALID;	
}

GnutellaPacket::~GnutellaPacket()
{
	if (packetBody)
		delete [] packetBody;
}

void GnutellaPacket::readHeaderFromString(unsigned char *verdomme)
{
	memcpy(header, verdomme, GNUTELLA_HEADER_LENGTH);

	switch(verdomme[HEADER_FUNCTION])
	{
		case	FUNC_INIT		:
		case	FUNC_INIT_RESP		:
		case	FUNC_PUSH		:
		case	FUNC_SEARCH		:
		case	FUNC_SEARCH_RESP	:
					break;
		default:
				verdomme[HEADER_FUNCTION]	= FUNC_INVALID;
				break;

	}
}

void GnutellaPacket::readBodyFromString(unsigned char *buf)
{
	if (packetBody)
		delete [] packetBody;

	if (getDataLength() > 0)
	{
		packetBody	= (unsigned char *)malloc(getDataLength());
		memcpy(packetBody, buf, getDataLength());
	}
	else
		packetBody	= 0;
}

int GnutellaPacket::mkForward(unsigned char maxTTL, unsigned char maxHops)
{
	if ((header[HEADER_HOPS] + header[HEADER_TTL]) > maxTTL)
	{
		if (header[HEADER_HOPS] < maxTTL)
			header[HEADER_TTL]	= maxTTL - header[HEADER_HOPS];
		else
			return 0;
	}

	if (header[HEADER_TTL] >= maxTTL)
		 header[HEADER_TTL]	= maxTTL;

	if (header[HEADER_TTL] > 1)
		header[HEADER_TTL]--;

	if (header[HEADER_HOPS] < 255)
		header[HEADER_HOPS]++;

	return (header[HEADER_TTL] && (header[HEADER_HOPS] <= maxHops));
}

unsigned int GnutellaPacket::setBody(char *newBody, unsigned int len)
{
	if (packetBody)
		delete [] packetBody;

	*(unsigned int *)&header[HEADER_DATALENGTH]	= len;
	if (newBody)
	{
		packetBody	= (unsigned char *)malloc(len);
		memcpy(packetBody, newBody, len);
	}
	else
	{
		packetBody			= NULL;
		*(unsigned int *)&header[HEADER_DATALENGTH]	= 0;
		if (len)
			return 0;
	}
	return len;
}

unsigned int GnutellaPacket::setBodyDirect(char *newBody, unsigned int len)
{
	if (packetBody)
		delete [] packetBody;

	*(unsigned int *)&header[HEADER_DATALENGTH]	= len;
	if ( (!((packetBody	= (unsigned char *)newBody))) && (len > 0) )
		return 0;
	else
		return len;
}

int GnutellaPacket::getBody(unsigned char *buf, unsigned int buflen)
{
	if (*(unsigned int *)&header[HEADER_DATALENGTH] > buflen)
		return -1;

	memcpy(buf, packetBody, *(unsigned int *)&header[HEADER_DATALENGTH]);

	return 	*(int *)&header[HEADER_DATALENGTH];
}

void GnutellaPacket::getHeader(unsigned char *buf)
{
	memcpy(buf, header, GNUTELLA_HEADER_LENGTH);
}

unsigned char GnutellaPacket::getType()
{
	return header[HEADER_FUNCTION];
}

void GnutellaPacket::setType(unsigned char newType)
{
	header[HEADER_FUNCTION]	= newType;
}

void GnutellaPacket::getMesgID(unsigned char *buf)
{
	memcpy((char *)buf, (char *)header, MESSAGE_ID_LENGTH);
}

void GnutellaPacket::setMesgID(unsigned char *buf)
{
	memcpy((char *)header, (char *)buf, MESSAGE_ID_LENGTH);
}

unsigned int GnutellaPacket::getDataLength()
{
	return *(unsigned int *)&header[HEADER_DATALENGTH];
}

int GnutellaPacket::getSourceConnection()
{
	return sourceConnection;
}

void GnutellaPacket::setSourceConnection(int newc)
{
	sourceConnection	= newc;
}

int GnutellaPacket::getTargetConnection()
{
	return targetConnection;
}

void GnutellaPacket::setTargetConnection(int newc)
{
	targetConnection	= newc;
}

void GnutellaPacket::GPinit(unsigned char defaultTTL)
{
	header[HEADER_FUNCTION]		= FUNC_INVALID;
	header[HEADER_TTL]		= defaultTTL;
	*(int *)&header[HEADER_DATALENGTH]	= 0;
	header[HEADER_HOPS]		= 0;
	packetBody			= 0;
}

int GnutellaPacket::isSane()
{
	switch(header[HEADER_FUNCTION])
	{
		case	FUNC_INIT		:
					if (getDataLength())
						return 0;
					if (packetBody)
						return 0;
					return 1;
		case	FUNC_INIT_RESP		:
					if (getDataLength() != INIT_RESP_BODY_SIZE)
						return 0;
					if (!packetBody)
						return 0;
					// port == 0 ?
					if (!(*(unsigned short *)packetBody))
						return 0;
					// IP addr == 0 ?
					if (!(*(unsigned int *)(packetBody + 2)))
						return 0;
					return 1;
		case	FUNC_PUSH		:
					if (getDataLength() != PUSH_BODY_SIZE)
						return 0;
					if (!packetBody)
						return 0;
					// IP addr == 0 ?
					if (!(*(unsigned int *)(packetBody + 20)))
						return 0;
					// port == 0 ?
					if (!(*(unsigned short *)(packetBody + 24)))
						return 0;
					return 1;
		case	FUNC_SEARCH		:
					if (	(getDataLength() < MIN_SEARCH_BODY_SIZE)	||
						(getDataLength() > MAX_SEARCH_BODY_SIZE)	)
						return 0;
					if (!packetBody)
						return 0;
					if (packetBody[getDataLength() - 1])
						return 0;
					return 1;
		case	FUNC_SEARCH_RESP	:
					if (	(getDataLength() < MIN_SEARCH_RESP_BODY_SIZE)	||
						(getDataLength() > MAX_SEARCH_RESP_BODY_SIZE)	)
						return 0;
					if (!packetBody)
						return 0;
					// port == 0 ?
					if (!(*(unsigned short *)(packetBody + 1)))
						return 0;
					// IP addr == 0 ?
					if (!(*(unsigned int *)(packetBody + 3)))
						return 0;
					return 1;
		default	: 		return 0;
	}
	return 0;
}

void GnutellaPacket::newMesgID()
{
	unsigned int 	seed;
	unsigned int 	i;

	sprintf((char *)header, "%d", (int)time(0));
	seed	= header[2] * header[3];
	srand(seed);
	i	= rand();
	memcpy(header + 8, (char *)&i, 4);
	i	= rand();
	memcpy(header + 12, (char *)&i, 4);
}

/*
 * DBElement inheritance
 */

int GnutellaPacket::equals(DBElement *gp)
{
	if (sourceConnection != ((GnutellaPacket *)gp)->getSourceConnection())
		return 0;

	return (!strncmp((char *)header, (char *)((GnutellaPacket *)gp)->getMesgIDp(),
			 MESSAGE_ID_LENGTH));
}

int GnutellaPacket::lessThan(DBElement *gp)
{
	if (sourceConnection > ((GnutellaPacket *)gp)->getSourceConnection())
		return 0;
	if (sourceConnection < ((GnutellaPacket *)gp)->getSourceConnection())
		return 1;

	return (strncmp((char *)header, (char *)((GnutellaPacket *)gp)->getMesgIDp(),
			 MESSAGE_ID_LENGTH) < 0);
}

int GnutellaPacket::greaterThan(DBElement *gp)
{
	if (sourceConnection < ((GnutellaPacket *)gp)->getSourceConnection())
		return 0;
	if (sourceConnection > ((GnutellaPacket *)gp)->getSourceConnection())
		return 1;

	return (strncmp((char *)header, (char *)((GnutellaPacket *)gp)->getMesgIDp(),
			 MESSAGE_ID_LENGTH) > 0);
}

unsigned int GnutellaPacket::hashValue()
{
	return (	(*(unsigned int *)header)	+
			(*(unsigned int *)(header + 5))	);
}

/*
 * FAME
 */

char *GnutellaPacket::getBodyp()
{
	return (char *)packetBody;
}

char *GnutellaPacket::getHeaderp()
{
	return (char *)header;
}

char *GnutellaPacket::getMesgIDp()
{
	return (char *)header;
}

