/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#define GNUTELLA_HEADER_LENGTH		23

#define HEADER_FUNCTION			16
#define HEADER_TTL			17
#define HEADER_HOPS			18
#define HEADER_DATALENGTH		19

#define FUNC_INIT			0x00
#define FUNC_INIT_RESP			0x01
#define FUNC_PUSH            		0x40
#define FUNC_SEARCH          		0x80
#define FUNC_SEARCH_RESP     		0x81
#define FUNC_INVALID			0xff

#define INIT_BODY_SIZE			0
#define INIT_RESP_BODY_SIZE		14
#define PUSH_BODY_SIZE			26
#define MIN_SEARCH_BODY_SIZE		3	// 2 + 1 = speed + zero
#define MAX_SEARCH_BODY_SIZE		256	// 255 + 1 = filename + zero
#define MIN_SEARCH_RESP_BODY_SIZE	27	// 0 records
#define MAX_SEARCH_RESP_BODY_SIZE	67624	// See Connection.h
#define MAX_BODY_LENGTH			(MAX_SEARCH_RESP_BODY_SIZE + 1)

class GnutellaPacket : public DBElement
{
	protected:
		unsigned char	header[GNUTELLA_HEADER_LENGTH];	
		unsigned char	*packetBody;
		int		sourceConnection;
		int		targetConnection;
		void		newMesgID();

	public:
		GnutellaPacket(unsigned char maxTTL);

		GnutellaPacket(int newCon, unsigned char maxTTL);

		GnutellaPacket(int newPacketType, char * newBody,
				int len, unsigned char defaultTTL);

		/*
		 * This constructor makes a packet that maches
		 * 'ask' as a response.
		 */
		GnutellaPacket(GnutellaPacket * ask);

		~GnutellaPacket();

		/*
		 * FALSE : Do not forward
		 */
		int		mkForward(unsigned char maxTTL, unsigned char maxHops);

		void		readBodyFromString(unsigned char *buf);

		void		readHeaderFromString(unsigned char *buf);

		/*
		 * setBody malloc's a new char [] and copies newBody
		 * setBodyDirect add's newBody as its body. DON'T MODIFY it after this!
		 */
		unsigned int	setBody(char *newBody, unsigned int len);

		unsigned int	setBodyDirect(char *newBody, unsigned int len);

		/*
		 * return values : size of body,
		 * -1 if size of body > buflen
		 */
		int		getBody(unsigned char *buf, unsigned int buflen);

		void		getHeader(unsigned char *buf);

		unsigned char	getType();

		void		setType(unsigned char newType);

		int		getSourceConnection();

		void		setSourceConnection(int newc);

		int		getTargetConnection();

		void		setTargetConnection(int newc);

		void		getMesgID(unsigned char *buf);

		char		*getMesgIDp();

		void		setMesgID(unsigned char *buf);

		unsigned int	getDataLength();

		void		GPinit(unsigned char defaultTTL);

		int		isSane();

		/*
		 * FAME
		 */

		char		*getBodyp();

		char		*getHeaderp();

		/*
		 * DBElement inheritance
		 */

		int		equals(DBElement *gp);

		int		lessThan(DBElement *gp);

		int		greaterThan(DBElement *gp);

		unsigned int	hashValue();
};
