/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#if (( OS == LINUX ) || ( OS == FREEBSD ))
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Constants.h"
#include "DBElement.h"
#include "DB.h"
#include "Socket.h"
#include "Buffer.h"
#include "Connection.h"
#include "FSEntry.h"
#include "Path.h"
#include "File.h"
#include "Transfer.h"
#include "Upload.h"
#include "HTTPServer.h"

HTTPServer::HTTPServer()
{
}

Upload *HTTPServer::serve(Connection *c, DB *files)
{
	Upload			*u;
	char			request[512];
	int			fid;
	int			bytesRead;
	int			offset;
	unsigned int 	len;
	int			i;
	char			fn[FILE_NAME_LENGTH];
	char			*finder;
	File			*f;
	Path			*fpath;
	struct sockaddr_in	sa;
 	
 	if (c->getState() != CON_STATE_CONNECTED)
	{
		delete c;
		return 0;
	}

	if (!files)
	{
		SEND404(c);
		delete c;
		return 0;
	}

	if ((bytesRead = c->peekInBuffer(request, 510)) < 12)
 		return 0;

	request[bytesRead]	= 0;
	request[511]		= 0;

	if (	(!strstr(request, "\r\n\r\n"))	||
		(!strstr(request, "GET /get/"))	||
		(!strstr(request, "HTTP/1."))	)
		return 0;

	finder	= &(request[0]);
	finder	+= strlen("GET /get/");
	fid	= atoi(finder);
	while (*finder != '/')
		finder++;
	finder++;
	for (i = 0; (	(i < FILE_NAME_LENGTH -1)	&&
			(finder[i]) 			&& 
			strncmp(" HTTP/1.", finder + i, strlen(" HTTP/1."))); 
								i++)
		fn[i]	= finder[i];
	fn[i]	= 0;
	if (!(f = (File *)files->findElement(fid)))
	{
		SEND404(c);
		c->close();
		delete c;
		return 0;
	}

	if (strncmp(fn, f->getFileNamep(), FILE_NAME_LENGTH))
	{
		SEND404(c);
		c->close();
		delete c;
		return 0;
	}

	/*
	 * Everything seems fine. Check for a content-range / offset
	 */
	if ((finder = strstr(request, "Content-Range:")))
	{
		finder	+= strlen("Content-Range:");
		while (*finder == ' ')
			finder++;
		offset	= atoi(finder);
	}
	else
		offset	= 0;

	/*
	 * Skip the HTTP header
	 */
	if (!(finder = strstr(request, "\n")))
	{
		c->close();
		delete c;
		return 0;
	}
	finder	+= strlen("\n");
	c->skipBytes((int)(finder - request));

	/*
	 * Create an Upload and store it.
	 */
	fpath	= f->getPath();
	len	= sizeof(sa);
	getsockname(c->getConnection(), (struct sockaddr *)&sa, &len);
	u	= new Upload(c, fid, f->getFileNamep(),
				fpath->getPathp(), offset, (char *)"123456789abcdef");
	u->setType(CON_TYPE_UPLOAD);
	if (!u->Start(LOCAL_READ))
	{
		delete u;
		delete c;
		return 0;
	}
	
	u->setState(CON_STATE_CONNECTED);
	delete c;
	
	return u;
}
#endif
