/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
 
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <dirent.h>

#include "Config.h"
#include "Constants.h"
#include "ControlCentre.h"
#include "DBElement.h"
#include "DB.h"
#include "GnutellaPacket.h"
#include "SearchResult.h"
#include "Packet.h"
#include "Mesg.h"
#include "Socket.h"
#include "Buffer.h"
#include "Connection.h"
#include "GnutellaConnection.h"
#include "Host.h"
#include "FSEntry.h"
#include "Path.h"
#include "File.h"
#include "Inet.h"
#include "Transfer.h"
#include "Upload.h"
#include "Download.h"
#include "TransferStatus.h"
#include "LocalResult.h"
#include "HTTPServer.h"
#include "FSScanner.h"
#include "FServer.h"
#include "HostCatcher.h"
#include "Query.h"
#include "Monitor.h"
#include "HTTPClient.h"
#include "Downloader.h"
#include "SearchManager.h"
#include "Networker.h"
#include "GnutellaRouter.h"
#include "Backend.h"
#ifndef ROUTER
#include "Commander.h"
#endif
#include "Hagelslag.h"

int main(int argc, char **argv)
{
	Hagelslag	h;

	if (argc > 1)
		if (!strncmp(argv[1], "-v", 2))
		{
			printf("%s\r\n", VERSION);
			return 0;
		}

	// prevent crashes from people who cancel during a send
	signal(SIGPIPE, SIG_IGN);

	return h.mainloop(argc, argv);
}

Hagelslag::Hagelslag()
{
	inputPtr	= 0;
}

int Hagelslag::mainloop(int argc, char **argv)
{
#ifndef ROUTER
	Commander	*cmdr;
	char		buf[COMMAND_BUF_SIZE];
	char		outbuf[OUT_SIZE];
#endif
	Backend		*be;
#ifdef ROUTER
	char		fileName[FILE_NAME_LENGTH];
	int		loopCount;
#endif

	checkDir();
	MCC	= startup();

#ifndef ROUTER
	cmdr	= new Commander(MCC);
#endif

#ifdef DAEMON
	readDeamonPort(argc, argv);
#endif

	be	= new Backend(MCC);

	setClientID(be);
	readHosts(be);
	loadShares(be);
	be->startListening();

#ifdef DAEMON
	openInput();
#endif

#ifdef ROUTER
	loopCount	= 0;
	sprintf(fileName, "%s/.hagelslag/hagelslag.hosts",
						getenv("HOME"));
#else
	toUser(PROMPT);
#endif


#ifndef ROUTER
	while(cmdr->keepRunning())
	{
		be->pollGnutellaNet();

		if (getInput(buf))
		{
#ifdef DAEMON
			if (!strncmp("holdon", buf, strlen("holdon")))
			{
				close(sock);
				connected = 0;
				continue;
			}
#endif
			cmdr->doCommand(buf, be, outbuf);
			toUser(outbuf);
			if (cmdr->keepRunning() && !*MCC->getError())
				toUser(PROMPT);
		}
		if (MCC->getNotifyResults())
		{
			MCC->clearNotifyResults();
			toUser("\nSearch results received!!!\n");
			toUser(PROMPT);			
		}
		if (*(MCC->getError()))
		{
			toUser(MCC->getError());
			if (cmdr->keepRunning())
				toUser(PROMPT);			
			MCC->clearError();
		}
	}

#endif

#ifdef ROUTER
	while(1)
	{
		if ( !be->pollGnutellaNet() )
			sleep (1);

		if (loopCount++ > 5000)
		{
			be->writeHosts(fileName);
			loopCount	= 0;
		}
	}
#endif

	return 0;
}

void Hagelslag::setClientID(Backend *be)
{
	FILE	*f;
	char	cid[100];
	char	fileName[FILE_NAME_LENGTH];

	sprintf(fileName, "%s/.hagelslag/hagelslag.cid", getenv("HOME"));
	if (!(f	= fopen(fileName, "r")))
	{
		if (!(f = fopen(fileName, "w")))
		{
			MCC->addError("Unable to open ~/.hagelslag/"
					"hagelslag.cid!!!");
			exit (0);
		}
		be->newClientID(cid);
		cid[CLIENTID_LENGTH]	= 0;
		fprintf(f, "%s\n", cid);
	}
	else
		fscanf(f, "%s", cid);

	fclose(f);
	be->set_STR_PARAM(PARAM_CLIENTID, cid);
}

void Hagelslag::checkDir()
{
	char	dirName[FILE_NAME_LENGTH];
	sprintf(dirName, "%s/.hagelslag", getenv("HOME"));
	mkdir(dirName, 0777);
}

void Hagelslag::readHosts(Backend *be)
{
	FILE	*f;
	char	fileName[FILE_NAME_LENGTH];

	sprintf(fileName, "%s/.hagelslag/hagelslag.hosts",
					getenv("HOME"));
	if ( !(f = fopen(fileName, "r")) )
	{
		MCC->addError("opening ~/.hagelslag/hagelslag.hosts failed!");
		return;
	}

	while (!feof(f))
	{
		int	tries;
		char	host[25];
		char	cid[100];

		fscanf(f, "%s %d %s", host, &tries, cid);
		host[24]	= 0;
		if (!strcmp(host, ""))
			continue;
		if (!strncmp(cid, "CID:", strlen("CID:")))
		{
			int	i;

			/*
			 * Read ' '
			 */
			fscanf(f, "%c", cid);
			if (cid[0] != ' ')
			{
				fprintf(stderr, "Error : No ' ' after CID: !!!\n");
				return;
			}
			for(i = 0; i < CLIENTID_LENGTH; ++i)
				fscanf(f, "%c", cid + i);
			cid[i]	= 0;
			be->storeHost(host, tries, cid);
		}
		else
			be->storeHost(host, tries, 0);
		host[0]	= 0;
	}
	fclose(f);	
}

ControlCentre *Hagelslag::startup()
{
	char		param[256];
	int		value;
	char		svalue[256];
	FILE		*conf;
	ControlCentre	*MCC;
	char		fileName[FILE_NAME_LENGTH];

	MCC		= new ControlCentre();
	sprintf(fileName, "%s/.hagelslag/hagelslag.conf", getenv("HOME"));
	if (!((conf = fopen(fileName, "r"))))
		return MCC;

	param[0]	= 0;
	while(!feof(conf))
	{
		fscanf(conf, "%s ", param);

		if (!strcmp(param, "DOWNLOAD_DIR"))
		{
			fscanf(conf, "%s\n", svalue);
			MCC->set_DOWNLOAD_DIR(svalue);
			continue;
		}
		fscanf(conf, "%d\n", &value);
		if (!strcmp(param, "MAX_TTL"))
			MCC->set_MAX_TTL(value);
		else if (!strcmp(param, "DEFAULT_TTL"))
			MCC->set_DEFAULT_TTL(value);
		else if (!strcmp(param, "ROUTING_TABLE_SIZE"))
			MCC->set_ROUTING_TABLE_SIZE(value);
		else if (!strcmp(param, "MAX_CONNECTIONS"))
			MCC->set_MAX_CONNECTIONS(value);
		else if (!strcmp(param, "MAX_RESULTS"))
			MCC->set_MAX_RESULTS(value);
		else if (!strcmp(param, "MAX_HOPS"))
			MCC->set_MAX_HOPS(value);
		else if (!strcmp(param, "MAX_PUSH_REQUESTS"))
			MCC->set_MAX_PUSH_REQUESTS(value);
		else if (!strcmp(param, "MAX_UPLOADS"))
			MCC->set_MAX_UPLOADS(value);
		else if (!strcmp(param, "MAX_DOWNLOADS"))
			MCC->set_MAX_DOWNLOADS(value);
		else if (!strcmp(param, "MAX_HOSTS"))
			MCC->set_MAX_HOSTS(value);
		else if (!strcmp(param, "PORT"))
			MCC->set_PORT(value);
		else if (!strcmp(param, "SPEED"))
			MCC->set_SPEED(value);
		else if (!strcmp(param, "MAX_MONITOR_LENGTH"))
			MCC->set_MAX_MONITOR_LENGTH(value);
		else if (!strcmp(param, "AUTO_CLEAR_COMPLETE"))
			MCC->set_AUTO_CLEAR_COMPLETE(value);
		else if (!strcmp(param, "AUTO_CONNECTIONS"))
			MCC->set_AUTO_CONNECTIONS(value);
		else if (!strcmp(param, "AUTO_PUSH"))
			MCC->set_AUTO_PUSH(value);
		else if (!strcmp(param, "AUTO_CLEAR_RESULTS"))
			MCC->set_AUTO_CLEAR_RESULTS(value);
		else if (!strcmp(param, "MAX_HOST_CON_FAIL"))
			MCC->set_MAX_HOST_CON_FAIL(value);
		else if (!strcmp(param, "AUTO_SAVE_HOSTS"))
			MCC->set_AUTO_SAVE_HOSTS(value);
		else if (!strcmp(param, "MAX_CON_ERRORS"))
			MCC->set_MAX_CON_ERRORS(value);
		else if (!strcmp(param, "MAX_FILES_SHARED"))
			MCC->set_MAX_FILES_SHARED(value);
		else if (!strcmp(param, "IGNORE_LOCAL_IP"))
			MCC->set_IGNORE_LOCAL_IP(value);
		else if (!strcmp(param, "NOTIFY_RESULTS"))
			MCC->set_NOTIFY_RESULTS(value);
		else if (!strcmp(param, "CACHE_ENTRIES"))
			MCC->set_CACHE_ENTRIES(value);
#ifdef DAEMON
		else if (!strcmp(param, "DAEMON_PORT"))
			MCC->set_DAEMON_PORT(value);
#endif
/*
		else if (!strcmp(param, ""))
			MCC->set_(value);
*/
		else
		{
			MCC->addError("Unknown option in config file");
		}

	}
	fclose (conf);

	return MCC;
}

#ifndef ROUTER

void Hagelslag::toUser(char *buf)
{
#ifdef CONSOLE
	fprintf(stderr, buf);
#endif
#ifdef DAEMON
	if (connected)
#ifdef MSG_NOSIGNAL
		send(sock, buf, strlen(buf), MSG_NOSIGNAL);
#else
		send(sock, buf, strlen(buf), 0);
#endif
#endif
}

#ifdef DAEMON
void Hagelslag::openInput()
{
	const	int		reuseaddr	= 1;
	struct  sockaddr_in	sa;

	sa.sin_family		= AF_INET;
	sa.sin_addr.s_addr	= htonl(INADDR_ANY);
	sa.sin_port		= htons(MCC->get_DAEMON_PORT());

	if ((servSock = socket(PF_INET, SOCK_STREAM, 0)) < 0)
	{
		fprintf(stderr, "Error : opening socket in openInput().\n");
		return;
	}
	setsockopt(servSock, SOL_SOCKET, SO_REUSEADDR, &reuseaddr,
			sizeof(reuseaddr));
	if (bind(servSock, (struct sockaddr *)&sa, sizeof(sa)) < 0)
	{
		fprintf(stderr, "Error : binding socket in openInput().\n");
		close(servSock);
		return;
	}
	listen(servSock, 0);
	connected	= 0;
}
#endif

int Hagelslag::getInput(char *buf)
{
	fd_set		in_set;	
	struct timeval	tm;

	buf[0]	= 0;

#ifdef CONSOLE
#define IN_CHANNEL 0
#endif
#ifdef DAEMON
#define IN_CHANNEL sock
	struct sockaddr_in	sa;
	unsigned int		len;
	int			br;

	if (!connected)
	{
		tm.tv_sec	= 0;
		tm.tv_usec	= 1;
		FD_ZERO(&in_set);
		FD_SET(servSock, &in_set);
		if (select(servSock + 1, &in_set, NULL, NULL, &tm) <= 0)
			return 0;
		len	= sizeof(sa);
		sock	= accept(servSock, (struct sockaddr *)&sa, &len);
		if (sock != -1)
			connected	= 1;
		toUser(PROMPT);
		return 0;
	}
#endif
	tm.tv_sec	= 0;
	tm.tv_usec	= 1;
	FD_ZERO(&in_set);
	FD_SET(IN_CHANNEL, &in_set);

	if ( select(IN_CHANNEL + 1, &in_set, 0, 0, &tm) <= 0)
		return 0;
#ifdef CONSOLE
	if (!fgets(buf, COMMAND_BUF_SIZE - 1, stdin))
	{
		strcpy(buf, "quit");
		return 1;
	}
	buf [strlen(buf) - 1] = 0;
#endif
#ifdef DAEMON
	if ((br = recv(sock, buf + inputPtr, COMMAND_BUF_SIZE - 1 - inputPtr, 0))
								 == -1)
	{
		close(sock);
		connected	= 0;
		return 0;
	}
	if (memchr(buf, '\n', COMMAND_BUF_SIZE))
	{
		*((char *)memchr(buf, '\n', COMMAND_BUF_SIZE))	= 0;
		if (memchr(buf, '\r', COMMAND_BUF_SIZE))
			*((char *)memchr(buf, '\r', COMMAND_BUF_SIZE))	= 0;
		inputPtr	= 0;
	}
	else
	{
		inputPtr	+= br;
		return 0;
	}
#endif

	return 1;
}

#endif

void Hagelslag::loadShares(Backend *be)
{
	char	name[FILE_NAME_LENGTH];
	FILE	*f;

	sprintf(name, "%s/.hagelslag/hagelslag.shares", getenv("HOME"));
	if (!((f = fopen(name, "r"))))
		return;

	name[0]	= 0;
	while (!feof(f))
	{
		if (fscanf(f, "%s\n", name) > 0)
		{
			if (strcmp(name, ""))
				be->addPath(name);
		}
	}
	fclose(f);
	be->scanDirs();
}

#ifdef DAEMON

int Hagelslag::readDeamonPort(int argc, char **argv)
{
	int	pt;

	if (argc <= 2)
		return 0;

	if (argc == 3)
	{
		if (strcmp(argv[1], "-p"))
		{
			fprintf(stderr, "Error : Unknown option %s\r\n",
							argv[1]);
			return 0;
		}
		pt	= atoi(argv[2]);

		if ((pt < 1) || (pt > 65535))
		{
			fprintf(stderr, "Error : invalid port %d\n", pt);
			return 0;
		}
		MCC->set_DAEMON_PORT(pt);
		return pt;
	}

	return 0;
}

#endif
