/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdio.h>

#include "Constants.h"
#include "DBElement.h"
#include "Inet.h"
#include "Host.h"

Host::Host()
{
	int		i;

	port		= 0;
	fileCount	= 0;
	fileSize	= 0;
	real		= 1;
	netIP		= 0;
	for (i = 0; i < CLIENTID_LENGTH; i++)
		clientID[i]	= 0;
}

Host::Host(unsigned long newNetIP, unsigned short newPort)
{
	int		i;

	netIP		= newNetIP;
	port		= newPort;
	fileCount	= 0;
	fileSize	= 0;
	real		= 1;
	tries		= 0;
	for (i = 0; i < CLIENTID_LENGTH; i++)
		clientID[i]	= 0;
}

Host::Host(unsigned long newNetIP, unsigned short newPort, int fc, int fs)
{
	int		i;

	netIP		= newNetIP;
	port		= newPort;
 	fileCount	= fc;
	fileSize	= fs;
	real		= 1;
	tries		= 0;
	for (i = 0; i < CLIENTID_LENGTH; i++)
		clientID[i]	= 0;
}

Host::Host(unsigned long newNetIP, unsigned short newPort, char *cid)
{
	netIP		= newNetIP;
	port		= newPort;
	fileCount	= 0;
	fileSize	= 0;
	real		= 1;
	tries		= 0;
	memcpy(clientID, cid, CLIENTID_LENGTH);
}

Host::Host(unsigned long newNetIP, unsigned short newPort, char *cid,
			int fc,int fs)
{
	netIP		= newNetIP;
	port		= newPort;
 	fileCount	= fc;
	fileSize	= fs;
	real		= 1;
	tries		= 0;
	memcpy(clientID, cid, CLIENTID_LENGTH);
}

Host::Host(char *dotted, char *cid)
{
	int		i;
	unsigned int	a, b, c, d;

	sscanf(dotted, "%u.%u.%u.%u:%hu", &a, &b, &c, &d, &port);
	real		= 1;
	tries		= 0;
	if (strchr(dotted, ':'))
	{
		*strchr(dotted, ':')	= 0;
		netIP	= Inet::dottedToIP(dotted);
	}
	else
		netIP	= 0;
	if (cid)
		memcpy(clientID, cid, CLIENTID_LENGTH);
	else
		for (i = 0; i < CLIENTID_LENGTH; i++)
			clientID[i]	= 0;
}

Host::~Host()
{
}

int Host::hasClientID()
{
	int	i;

	for (i = 0; i < CLIENTID_LENGTH; i++)
		if (clientID[i])
			return 1;

	return 0;
}

void Host::incTries()
{
	tries++;
}

void Host::setTries(int t)
{
	tries	= t;
}

int Host::getTries()
{
	return tries;
}

void Host::getClientID(char *buf)
{
	memcpy(buf, clientID, CLIENTID_LENGTH);
}

void Host::setClientID(char *buf)
{
	memcpy(clientID, buf, CLIENTID_LENGTH);
}

int Host::isReal()
{
	return real;
}

void Host::setReal(int nr)
{
	real	= nr;
}

unsigned long Host::getNetIP()
{
	return netIP;
}

unsigned short Host::getPort()
{
	return port;
}

void Host::getDotted(char *buf)
{
	Inet::IPPortToDotted(buf, netIP, port);
}

/*
 * DBElement Inheritance
 */

int Host::equals(DBElement *myHost)
{
	if (netIP != ((Host *)myHost)->getNetIP())
		return 0;
	if (port != (((Host *)myHost))->getPort())
		return 0;

	return 1;
}

int Host::lessThan(DBElement *myHost)
{
	if (netIP < ((Host *)myHost)->getNetIP())
		return 1;
	if (netIP > ((Host *)myHost)->getNetIP())
		return 0;

	if (port < ((Host *)myHost)->getPort())
		return 1;
	if (port > ((Host *)myHost)->getPort())
		return 0;

	return 0;
}

int Host::greaterThan(DBElement *myHost)
{
	if (netIP > ((Host *)myHost)->getNetIP())
		return 1;
	if (netIP < ((Host *)myHost)->getNetIP())
		return 0;

	if (port > ((Host *)myHost)->getPort())
		return 1;
	if (port < ((Host *)myHost)->getPort())
		return 0;

	return 0;
}

unsigned int Host::hashValue()
{
	return netIP;
}

/*
 * FAME
 */

char *Host::getClientIDp()
{
	return (char *)clientID;
}
