/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdio.h>
#include <time.h>

#include "Constants.h"
#include "ControlCentre.h"
#include "DBElement.h"
#include "DB.h"
#include "Socket.h"
#include "Buffer.h"
#include "GnutellaPacket.h"
#include "Connection.h"
#include "Host.h"
#include "HostCatcher.h"

HostCatcher::HostCatcher(ControlCentre *newMCC)
{
	MCC	= newMCC;
}

int HostCatcher::catchHost(GnutellaPacket *p, DB *hosts)
{
	char		GClientID[CLIENTID_LENGTH];
	int		hIndex;
	short		port;
	Host		*h;
	long		hostIP;

	switch (p->getType())
	{
		case FUNC_INIT_RESP	:
				if (hosts->isFull())
					return ENTRY_INVALID;

				h	= new Host(*((unsigned long *)(p->getBodyp() + 2)),
							*((unsigned long *)(p->getBodyp())));
				if (!isValidHost(h))
				{
					delete h;
					return ENTRY_INVALID;
				}
				if (((hIndex = hosts->storeElement(h))) == ENTRY_INVALID)
					delete h;
				else
					hostsCaught++;
				return hIndex;
		case FUNC_SEARCH_RESP	:
			//		PARAMETER	BODY + OFFSET	RANGE
			//	      |---------------|--------------|-------|
				memcpy(&port,		p->getBodyp() + 1, 	2);
				memcpy(&hostIP, 	p->getBodyp() + 3, 	4);
			        memcpy(GClientID, (p->getBodyp() + p->getDataLength() - CLIENTID_LENGTH),
						CLIENTID_LENGTH);
				h	= new Host(hostIP, port, GClientID);
				if (!isValidHost(h))
				{
					delete h;
					return ENTRY_INVALID;
				}
				if ((hIndex = hosts->containsElement(h)) == ENTRY_INVALID)
				{
					if (hosts->isFull())
					{
						delete h;
						return ENTRY_INVALID;
					}

					if ((hIndex = hosts->storeElement(h)) == ENTRY_INVALID)
						delete h;
					else
						hostsCaught++;
				}
				else
				{
					delete h;
					h	= (Host *)hosts->findElement(hIndex);
					if (!h->hasClientID())
						h->setClientID(GClientID);
				}
				return hIndex;
		default			: break;
	}
	return ENTRY_INVALID;
}

int HostCatcher::isValidHost(Host *h)
{
	char		dotted[DOTTED_LENGTH];

	if (!h->isReal())
		return 0;

	if (MCC->get_IGNORE_LOCAL_IP())
	{
		unsigned char	c;
		h->getDotted(dotted);

		/*
		 * The list of ignorable networks was provided by Brian Macke
		 */

		// Never header of anything starting with 0
		if (!strncmp(dotted, "0.", strlen("0.")))
			return 0;

		// Class A localnet
		if (!strncmp(dotted, "10.", strlen("10.")))
			return 0;

		// Class B localnet
		if (!strncmp(dotted, "172", strlen("172")))
		{
			sscanf(dotted, "172.%c.", &c);

			if ( (c > 16) && (c < 32) )
				return 0;
			return 1;
		}

		// Class C localnet
		if (!strncmp(dotted, "192.168.", strlen("192.168.")))
			return 0;

		// loopback
		if (!strncmp(dotted, "127", strlen("127")))
			return 0;

		// Local Link networks
		if (!strncmp(dotted, "169.254.", strlen("169.254.")))
			return 0;

		// ExampleNet
		if (!strncmp(dotted, "192.0.2.", strlen("192.0.2.")))
			return 0;

/*
		if (!strncmp(dotted, "", strlen("")))
			return 0;
*/
	}

	return 1;
}

unsigned int HostCatcher::getHostsCaught()
{
	return hostsCaught;
}
