/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "Config.h"

#if ( (OS == LINUX) || (OS == FREEBSD) )
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#endif

#include "Inet.h"

#if ( (OS == LINUX) || (OS == FREEBSD) )
unsigned long int Inet::dottedToIP(char *dotted)
{
	struct in_addr	inp;
	
	if (inet_aton(dotted, &inp))
		return inp.s_addr;
	
	return 0;
}

int Inet::IPPortToDotted(char *buf, unsigned long int netIP, unsigned short port)
{
	struct in_addr	ia;

	ia.s_addr	= netIP;
	sprintf(buf, "%s:%u", inet_ntoa(ia), port);
	return 1;
}

int Inet::IPToDotted(char *buf, unsigned long int netIP)
{
	struct in_addr	ia;

	ia.s_addr	= netIP;
	sprintf(buf, "%s", inet_ntoa(ia));
	return 1;
}

unsigned long int Inet::NSLookup(char *hostname)
{
	struct hostent	*h;

	if ((h = gethostbyname(hostname)))
		return *(unsigned long int *)(h->h_addr_list[0]);
	
	return 0;
}

int Inet::getHostName(char *host, int hlen)
{
	return (gethostname(host, hlen) != -1);
}
#else
#if (OS == WIN32)
#warning Inet : compiling for win32
unsigned long int Inet::dottedToIP(char *dotted)
{
}

int Inet::IPPortToDotted(char *buf, unsigned long int netIP, unsigned short port)
{
}

int Inet::IPToDotted(char *buf, unsigned long int netIP)
{
}

unsigned long int Inet::NSLookup(char *hostname)
{
}

int Inet::getHostName(char *host, int hlen)
{
}
#endif
#endif
