#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "Constants.h"
#include "DBElement.h"
#include "DB.h"
#include "GnutellaPacket.h"
#include "SearchResult.h"
#include "FSEntry.h"
#include "Path.h"
#include "File.h"
#include "Packet.h"

Packet::Packet(GnutellaPacket *p) : GnutellaPacket(p)
{
}

Packet::Packet(unsigned char defaultTTL) :
			GnutellaPacket(defaultTTL)
{
}

void Packet::reset(unsigned char defaultTTL)
{
	if (packetBody)
		delete [] packetBody;

	GPinit(defaultTTL);
}

int Packet::mkInitPacket(unsigned char defaultTTL)
{
	reset(defaultTTL);
	header[HEADER_FUNCTION]	= FUNC_INIT;
	return 1;
}

int Packet::mkInitRespPacket(short port, int ip, int count, int filesSize, unsigned char defaultTTL)
{
	reset(defaultTTL);
	header[HEADER_FUNCTION]				= FUNC_INIT_RESP;
	*(unsigned int *)&header[HEADER_DATALENGTH]	= INIT_RESP_BODY_SIZE;
	packetBody		= (unsigned char *)malloc(INIT_RESP_BODY_SIZE);
	if (!packetBody)
	{
		fprintf(stderr, "Error : mkInitRespPacket : malloc() failed!!!\n");
		return 0;
	}
	memcpy(packetBody, 	(char *)&port,	 	2);
	memcpy(packetBody + 2, 	(char *)&ip, 		4);
	memcpy(packetBody + 6, 	(char *)&count, 	4);
	memcpy(packetBody + 10, (char *)&filesSize,	4);
	return 1;
}

int Packet::mkPushPacket(char *CID, int fix, int ip, short port, unsigned char defaultTTL)
{
	reset(defaultTTL);
	header[HEADER_FUNCTION]			= FUNC_PUSH;
	*(int *)&header[HEADER_DATALENGTH]	= PUSH_BODY_SIZE;
	packetBody	= (unsigned char *)malloc(PUSH_BODY_SIZE);
	memcpy(packetBody, CID, CLIENTID_LENGTH);
	memcpy((packetBody + 16), &fix, 4);
	memcpy((packetBody + 20), &ip, 4);
	memcpy((packetBody + 24), &port, 2);

	return 1;
}

int Packet::mkSearchPacket(short speed, char *query, unsigned char defaultTTL)
{
 	int	sz;

	newMesgID();
	reset(defaultTTL);
	header[HEADER_FUNCTION]	= FUNC_SEARCH;
	sz	= strlen(query) + 1;
	if (sz	> FILE_NAME_LENGTH - 2)
		sz	= FILE_NAME_LENGTH - 2;
	*(int *)&header[HEADER_DATALENGTH]	= sz + 2;
	packetBody	= (unsigned char *)malloc(sz + 2);
	memcpy(packetBody, &speed, 2);
	memcpy(packetBody + 2, query, sz);
	packetBody[ sz + 1 ]	= 0;

	return 1;
}

int Packet::mkSearchRespPacket(char *CID, short port, int ip, int spd, unsigned char defaultTTL)
{
	reset(defaultTTL);
	header[HEADER_FUNCTION]	= FUNC_SEARCH_RESP;
	*(int *)&header[HEADER_DATALENGTH]	= MIN_SEARCH_RESP_BODY_SIZE;
	packetBody	= (unsigned char *)malloc(MIN_SEARCH_RESP_BODY_SIZE);

	/*
	 * Bugfix provided by Nate Riffe
	 * Corrected incorrect host speed in response, by writing CID
	 * to (packetBody + 11) in stead of to (packetBody + 8)
	 */
	packetBody[0]	= 0;
	memcpy(packetBody + 1, &port, 2);
	memcpy(packetBody + 3, &ip, 4);
	memcpy(packetBody + 7, &spd, 4);
	memcpy(packetBody + 11, CID, CLIENTID_LENGTH);

	return 1;
}

int Packet::setSearchRespIP(unsigned int ip)
{
	memcpy(packetBody + 3, &ip, 4);
	return 1;
}

/*
 * Most definately the most annoying method to write.
 */
int Packet::addToSRPacket(int index, int size, char *name)
{
	int		namelen;

	if (	(header[HEADER_FUNCTION] != FUNC_SEARCH_RESP)	||
		(!packetBody)	)
		return 0;

	if ((unsigned char)packetBody[0] == 255)
		return 0;

	// Make sure the name is DOUBLE-NULL Termintated.
	name[FILE_NAME_LENGTH - 1] 		= 0;
	namelen	= strlen(name) + 2;
	if (namelen >= FILE_NAME_LENGTH)
	{
		name[FILE_NAME_LENGTH - 1] 	= 0;
		name[FILE_NAME_LENGTH - 2] 	= 0;
		namelen				= FILE_NAME_LENGTH;
	}
	else
		name[namelen - 1]		= 0;

	packetBody	= (unsigned char *)realloc(packetBody,
				*(int *)&header[HEADER_DATALENGTH] + namelen + 8);
	if (!packetBody)
	{
		fprintf(stderr, "realloc Failed!!!\n");
		header[HEADER_FUNCTION]	= FUNC_INVALID;
		return 0;
	}

	// Copy clientID to end
	memcpy(packetBody + *(int *)&header[HEADER_DATALENGTH] - CLIENTID_LENGTH + namelen + 8,
		packetBody + *(int *)&header[HEADER_DATALENGTH] - CLIENTID_LENGTH,
		CLIENTID_LENGTH);

 	// continue copying, adding index and size
	memcpy(packetBody + *(int *)&header[HEADER_DATALENGTH] - CLIENTID_LENGTH,
			&index, 4);
	memcpy(packetBody + *(int *)&header[HEADER_DATALENGTH] - CLIENTID_LENGTH + 4,
			&size, 4);

	// continue copying, adding file name
	memcpy(packetBody + *(int *)&header[HEADER_DATALENGTH] - CLIENTID_LENGTH + 8,
					name, namelen);

	// increase datalength with the size of the new record
	*(int *)&header[HEADER_DATALENGTH]	= (	*(int *)&header[HEADER_DATALENGTH] +
						namelen + 8	);

	// increase num recs
	packetBody[0]++;

	return 1;
}

