/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

 #include "Config.h"

#include <string.h>
#include <stdio.h>

#if ((OS == LINUX) || (OS == FREEBSD))
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#endif

#include "Constants.h"
#include "DBElement.h"
#include "FSEntry.h"
#include "Path.h"
#include "File.h"

Path::Path() : FSEntry(FS_DIRECTORY)	
{
	path[0]	= 0;
	subdir	= 0;
}

Path::Path(char *newPath) : FSEntry(FS_DIRECTORY)
{
	strncpy(path, newPath, PATH_NAME_LENGTH - 1);
	path[PATH_NAME_LENGTH - 1]	= 0;
	subdir	= 0;
}

Path::Path(char *newPath, char newSubDir) : FSEntry(FS_DIRECTORY)
{
	strncpy(path, newPath, PATH_NAME_LENGTH - 1);
	path[PATH_NAME_LENGTH - 1]	= 0;
	subdir	= newSubDir;
}

void Path::setPath(char *newPath)
{
	strncpy(path, newPath, PATH_NAME_LENGTH - 1);
	path[PATH_NAME_LENGTH - 1]	= 0;
}

void Path::getPath(char *myPath)
{
	strncpy(myPath, path, PATH_NAME_LENGTH - 1);
}

int Path::isSubdir()
{
	return subdir;
}

void Path::setSubdir(int sd)
{
	subdir	= sd;
}

#if ((OS == LINUX) || (OS == FREEBSD))
FSEntry *Path::list()
{
	DIR		*dp;
	struct dirent	*de;
	FSEntry		*first;
	
	if (!((dp = opendir(path))))
		return 0;
	
	first	= 0;
	while ((de = readdir(dp)))
	{
		struct stat	fst;
		File		*f;
		Path		*p;
		char		fname[PATH_NAME_LENGTH];
		
		if ( (!strcmp(de->d_name, ".")) || (!strcmp(de->d_name, "..")) )
			continue;
		
		sprintf(fname, "%s/%s",  path, de->d_name);
		if (stat(fname, &fst) == -1)
			continue;
			
		/*
		 * alternative way to detect directory : opendir() == -1 ? it 's a file : it's a dir
		 */
		if (S_ISDIR(fst.st_mode))
		{
			p	= new Path(fname, 1);
			if (first)
				p->setNext(first);
			first	= p;
			continue;
		}
		
		/*
		 * It's a file
		 */
		f       = new File((int)fst.st_size, de->d_name,  this);
		if (first)
			f->setNext(first);
		first	= f;
	}
	
	return first;
}
#else
FSEntry *Path::list()
{
#warning WIN32:Path::list() not implemented!!!!
	return 0;
}
#endif

/*
 * FAME
 */

char *Path::getPathp()
{
	return (char *)path;
}

/*
 * DBElement inheritance
 */

int Path::equals(DBElement *myPath)
{
	return (!strncmp(path, ((Path *)myPath)->getPathp(), PATH_NAME_LENGTH));
}

int Path::lessThan(DBElement *myPath)
{
	return (strncmp(path, ((Path *)myPath)->getPathp(), PATH_NAME_LENGTH) < 0);
}

int Path::greaterThan(DBElement *myPath)
{
	return (strncmp(path, ((Path *)myPath)->getPathp(), PATH_NAME_LENGTH) > 0);
}
              
unsigned int Path::hashValue()
{
	int	hv;
	char	*p;
	
	p	= (char *)path;
	hv	= 0;
	
	while(*p++)
		hv	+= *p;

	return hv;
}
