/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <time.h>
#include <stdio.h>

#include "DBElement.h"
#include "RCPacket.h"
#include "RCPacketStorage.h"

RCPacketStorage::RCPacketStorage(int arraySize)
{
	int	i;
	
	size	= arraySize;
	iFree	= 0;
	packets	= 0;
	array	= new RCPacket* [size];
	
	for (i = 0; i < size; i++)
		array[i]	= 0;
}

RCPacketStorage::~RCPacketStorage()
{
	int	i;
	
	for (i = 0; i < size; i++)
	{
		RCPacket	*p;
		RCPacket	*tmp;
		
		p	= array[i];
		while (p)
		{
			tmp	= p;
			p	= p->getNext();
			delete	tmp;
		}
	}
	
	delete array;
	array	= 0;
}

int RCPacketStorage::findFreeIndex()
{
	int	i;
	
	if (iFree == PKTID_INVALID)
		return PKTID_INVALID;

	if (!array[iFree])
		return iFree++;

	i	= iFree;
	
	while (iFree < size)
	{
		if (!array[iFree])
			return iFree;
		iFree++;
	}
			
	iFree	= 0;
	while (iFree < i)
	{
		if (!array[iFree])
			return iFree;
		iFree++;
	}

	iFree	= PKTID_INVALID;
	
	return iFree;
}
		
RCPacket *RCPacketStorage::packetList(int pktID)
{
	if ( (pktID < 0) || (pktID >= size) )
	{
		fprintf(stderr, "ERROR : BUG : RCPacketStorage::packetList()"
				" : invalid index (%d)!\n", pktID);
		return 0;
	}

	return array[pktID];
}

int RCPacketStorage::storePacket(RCPacket *p)
{
	int		i;
	RCPacket	*l;
	
	i	= p->getPacketID();
	if ( (i < 0) || (i >= size) )
	{
		fprintf(stderr, "ERROR : BUG : RCPacketStorage::storePacket()"
				" : invalid index (%d)!\n", i);
		return 0;
	}
	for (l = array[i]; l; l = l->getNext())
		if (l->equals(p))
			return 0;
	
	p->setNext(array[i]);
	if (array[i])
		array[i]->setPrv(p);

	array[i]	= p;
	packets++;
	
	return 1;
}
		
int RCPacketStorage::deletePacket(RCPacket *p)
{
	int 		i;
	
	i	= p->getPacketID();
	
	if (p == array[i])
		array[i]	= p->getNext();
	else
		(p->getPrv())->setNext(p->getNext());
	
	if (p->getNext())
		(p->getNext())->setPrv(p->getPrv());

	if ((!array[i]) && (iFree == PKTID_INVALID))
		iFree	= 0;

	delete p;
	packets--;
	
	return 1;
}

int RCPacketStorage::getPackets()
{
	return packets;
}

int RCPacketStorage::deleteByPacketID(int pIndex)
{
	RCPacket	*p;
	RCPacket	*prv;

	if ( (pIndex < 0) || (pIndex >= size) )
	{
		fprintf(stderr, "ERROR : BUG : RCPacketStorage::deleteByPacketID()"
				" : invalid index (%d)!\n", pIndex);
		*(int*)0=0;
		return 0;
	}

	p	= array[pIndex];
	while(p)
	{
		prv	= p;
		p	= p->getNext();
		delete prv;
		packets--;
	}
	array[pIndex]	= 0;

	return 1;
}

int RCPacketStorage::integrityCheckByPacketID(int pIndex)
{
	RCPacket	*p;
	RCPacket	*prv;
	int		t;

	if ( (pIndex < 0) || (pIndex >= size) )
		return INTEGRITY_ERROR;

	p	= array[pIndex];
	t	= 0;
	while(p)
	{
		t++;
		prv	= p;
		p	= p->getNext();
		if (p == array[pIndex])
		{
			fprintf(stderr, "ERROR : INTEGRITY : CPacketStorage::integrityCheckByPacketID() :"
					"p == array[%d]!!!\r\n", pIndex);
			return INTEGRITY_ERROR;
		}
		if (p == prv)
		{
			fprintf(stderr, "ERROR : INTEGRITY : CPacketStorage::integrityCheckByPacketID() :"
					"p == prv (i == %d)!!!\r\n", pIndex);
			return INTEGRITY_ERROR;
		}
		if (prv->getPacketID() != pIndex)
		{
			fprintf(stderr, "ERROR : INTEGRITY : CPacketStorage::integrityCheckByPacketID() :"
					"packetIndex  %d differnt from pIndex %d!!!\r\n",
					p->getPacketID(), pIndex);
			return INTEGRITY_ERROR;
		}
	}

	return t;
}

int RCPacketStorage::integrityCheck()
{
	int	i;
	int	p;
	
	p	= 0;
	for (i = 0; i < size; i++)
	{
		int	res;
		
		if ((res = integrityCheckByPacketID(i)) == INTEGRITY_ERROR)
			return INTEGRITY_ERROR;
		p	+= res;
	}

	if (p != packets)
	{
		fprintf(stderr, "ERROR : INTEGRITY : CPacketStorage::integrityCheck() :"
				"p != packets!!!\r\n");
		return INTEGRITY_ERROR;
	}
	
	return p;
}

