/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "DBElement.h"
#include "RCRefList.h"
#include "RCPattern.h" 

RCPattern::RCPattern(char *newPattern)
{
	int	i;
	
	memcpy(pattern, newPattern, MAX_QUERY_SIZE);
	refCount	= 0;
	
	for (i = 0; i < RCPATTERN_REFARRAY_SIZE; i++)
		refListArray[i]	= 0;
}

RCPattern::~RCPattern()
{
	int	i;

	for (i = 0; i < RCPATTERN_REFARRAY_SIZE; i++)
		if (refListArray[i])
		{
			refListArray[i]->deleteRest();
			delete refListArray[i];
		}
}

/*
 * Doesn't check for doubles yet, it should do that.
 */
int RCPattern::addToRefList(int whichList, RCRefList *reference)
{
	if ( (whichList < RCPATTERN_REFLIST_MESGID) 	|| 
		(whichList >= RCPATTERN_REFARRAY_SIZE)	)
	{
		fprintf(stderr, "ERROR : BUG : RCPattern::addToRefList() : "
				"whichList (%d) invalid!!\r\n", whichList);
		return 0;
	}

	switch (whichList)
	{
		case RCPATTERN_REFLIST_MESGID	:
					reference->setNext(refListArray[RCPATTERN_REFLIST_MESGID]);
					refListArray[RCPATTERN_REFLIST_MESGID]	= reference;
					break;
		case RCPATTERN_REFLIST_PARAMETERS	:
					reference->setNext(refListArray[RCPATTERN_REFLIST_PARAMETERS]);
					refListArray[RCPATTERN_REFLIST_PARAMETERS]	= reference;
					break;
		case RCPATTERN_REFLIST_ENTRY	:
					reference->setNext(refListArray[RCPATTERN_REFLIST_ENTRY]);
					refListArray[RCPATTERN_REFLIST_ENTRY]	= reference;
					break;
		case RCPATTERN_REFLIST_SPAM	:
					reference->setNext(refListArray[RCPATTERN_REFLIST_SPAM]);
					refListArray[RCPATTERN_REFLIST_SPAM]	= reference;
					break;
		default	: return 0;
	}
	
	return 1;
}

int RCPattern::deleteFromRefList(int whichList, int index)
{
	RCRefList	*rl;
	RCRefList	*prv;

	if ( (whichList < RCPATTERN_REFLIST_MESGID) 	|| 
		(whichList >= RCPATTERN_REFARRAY_SIZE))
	{
		fprintf(stderr, "ERROR : BUG : RCPattern::deleteFromRefList() : "
				"whichList (%d) invalid!!\r\n", whichList);
		return 0;
	}

	prv	= 0;
	rl	= refListArray[whichList];
	while (rl)
	{
		if (rl->getIndex() == index)
		{
			if (prv)
				prv->setNext(rl->getNext());
			else
				refListArray[whichList]	= rl->getNext();

			delete rl;
			return 1;
		}
		prv	= rl;
		rl	= rl->getNext();
	}

	fprintf(stderr, "ERROR : BUG : RCPattern::deleteFromRefList() : "
			"index %d not found in list %d!!!\r\n", index,
			whichList);
	
	return 0;
}

RCRefList *RCPattern::getRefList(int whichList)
{
	if ( (whichList >= 0) && (whichList < RCPATTERN_REFARRAY_SIZE) )
		return refListArray[whichList];
	
	return 0;
}

char *RCPattern::getPatternp()
{
	return (char *)pattern;
}

int RCPattern::decRefCount()
{
	return --refCount;
}

int RCPattern::incRefCount()
{
	return ++refCount;
}

int RCPattern::getRefCount()
{
	return refCount;
}

int RCPattern::equals(DBElement *other)
{
	return !strncmp(pattern, ((RCPattern *)other)->getPatternp(), MAX_QUERY_SIZE);
}

int RCPattern::lessThan(DBElement *other)
{
	return (strncmp(pattern, ((RCPattern *)other)->getPatternp(), MAX_QUERY_SIZE) < 0);
}

int RCPattern::greaterThan(DBElement *other)
{
	return (strncmp(pattern, ((RCPattern *)other)->getPatternp(), MAX_QUERY_SIZE) > 0);
}

unsigned int RCPattern::hashValue()
{
	int	i;
	int	res;
	
	res	= 0;
	
	for (i = 0; pattern[i] && (i < MAX_QUERY_SIZE); i++)
		res	+= pattern[i];
		
	return res;
}


