/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#define SOCKET_CONNECT_SUCCESS		 1
#define SOCKET_CONNECT_FAILED		-1
#define SOCKET_CONNECT_IN_PROGRESS	 0

class Socket
{
	private:
		int			sock;
		unsigned long int	ip;
		unsigned short		port;

	public:
		/*
		 * ip	= 0;
		 * port	= 0;
		 * sock	= -1;
		 */
		Socket();

		Socket(Socket *s);

		/*
		 * ip	= 0;
		 * port	= 0;
		 * sock	= newSocket;
		 */
		Socket(int newSocket);

		/*
		 * closes current socket if connected
		 */
		~Socket();

		/*
		 * Syscall socket
		 */
		int	callSocket();

		/*
		 * Open assigns ip and port, then tries to open a
		 * NON-blocking tcp connection to that address.
		 * ip and port must be in network byte order.
		 * Returnvalues : 
		 * SOCKET_CONNECT_SUCCESS
		 * SOCKET_CONNECT_FAILED
		 * SOCKET_CONNECT_IN_PROGRESS
		 */
		int	callConnect(unsigned long int hostIp, unsigned short hostPort);

		/*
		 * return values :
		 * SOCKET_CONNECT_SUCCESS
		 * SOCKET_CONNECT_FAILED
		 * SOCKET_CONNECT_IN_PROGRESS
		 */
		int	testConnect();
		
		/*
		 * Returns Value :
		 *  1 : send would be succesfull
		 *  0 : send would NOT be succesfull
		 * -1 : Error occured
		 * usec is the number of millisecs to wait until end of test.
		 */
		int	testSend(int usec);

		/*
		 * Returns Value :
		 *  1 : recv would be succesfull
		 *  0 : recv would NOT be succesfull
		 * -1 : Error occured
		 * usec is the number of millisecs to wait until end of test.
		 */
		int	testRecv(int usec);

		/*
		 * Send and Recv return the number of bytes processed
		 * or a -1 on error
		 */
		int	callRecv(char *buf, int count);

		int	callSend(char *msg, int len);

		/*
		 * Peek in the receive buffer
		 * return number of char written to buf
		 */
		int	peek(char *buf, int count);

		/*
		 * Close the socket
		 */
		int	callClose();

		/*
		 * bind
		 */
		int	callBind(unsigned short port);

		/*
		 * Start listening on the socket
		 */
		int	callListen(int queue);

		/*
		 * accept connection. Returns new sock or NULL;
		 */
		Socket	*callAccept();

		/*
		 * callCloses current socket if connected
		 */
		int	setSocket(int newSock);

		/*
		 * Return the actual socket. // CHECK PORTABILITY
		 */
		int	getSock();

		/*
		 * ip addr and port are in in network byte order
		 */
		unsigned long int	getIp();

		unsigned short		getPort();

		void	setIp(unsigned long int newIp);

		void	setPort(unsigned short port);

		/*
		 * Prints the ip address in dotted (1.2.3.4) notation in buf
		 */
		void	getDottedIp(char *buf);
};
