/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "Socket.h"
#include "Constants.h"
#include "DBElement.h"
#include "Buffer.h"
#include "Connection.h"
#include "FSEntry.h"
#include "Path.h"
#include "File.h"
#include "Transfer.h"
#include "Upload.h"

Upload::Upload(Connection *c, int newFileID, char *newFileName,
		char *newDirName, int newOffset, char *cid)
	: Transfer(newFileID, newFileName, newDirName, newOffset, c)
{
	int	i;

        type            = CON_TYPE_UPLOAD;
	for (i = 0; i < 2 * CLIENTID_LENGTH; ++i)
	{
		unsigned char	idx;
		char		hex []	=  {	'0', '1', '2', '3', '4', '5', '6', '7',
					'8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

		if (i % 2)
			idx	= cid[i / 2] & 0xF;
		else
			idx	= (cid[i / 2] & 0xF0) / 0x10;

 		aacid[i]	= hex[ idx ];
	}
}

Upload::Upload(int newFileID, char *newFileName,
		char *newDirName, int newOffset, char *cid)
	: Transfer(newFileID, newFileName, newDirName, newOffset)
{
	int	i;

        type            = CON_TYPE_UPLOAD;
	for (i = 0; i < 2 * CLIENTID_LENGTH; ++i)
	{
		unsigned char	idx;
		char		hex []	=  {	'0', '1', '2', '3', '4', '5', '6', '7',
					'8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

		if (i % 2)
			idx	= cid[i / 2] & 0xF;
		else
			idx	= (cid[i / 2] & 0xF0) / 0x10;

 		aacid[i]	= hex[ idx ];
	}
}

int Upload::loginToPeer()
{
	char		buf[512];
	char		*bufp;
	Path		*p;
	File		*f;

	p	= new Path(dirName, 0);
	f	= new File(0, fileName, p);

	if (!f->statFile())
	{
		delete p;
		delete f;
		setState(CON_STATE_LOGIN_FAILED);
		return 0;
	}
	bytesTotal	= f->getSize();
	bytesLeft	= bytesTotal - offset;
	bytesDone	= 0;

	delete p;
	delete f;

	bufp	= (char *)buf;
       	switch (type)
	{
		case	CON_TYPE_PUSH_OUT	:
	       		sprintf(bufp, "GIV %d:", fileID);
			bufp	+= strlen(bufp);
			memcpy(bufp, aacid, (2 * CLIENTID_LENGTH));
			bufp	+= (2 * CLIENTID_LENGTH);
			sprintf(bufp, "/%s\n\n", fileName);
			bufp	+= strlen(bufp);
			// NO 'break;' here -- http login must follow!
		case	CON_TYPE_UPLOAD		:
		       	if (offset > 0)
		                sprintf(bufp,    "HTTP 200 OK\r\nContent-range: bytes="
	        	                        "%d-%d/%d\r\n\r\n", offset,
						bytesTotal - 1, bytesTotal);
		        else
		                sprintf(bufp,    "HTTP 200 OK\r\nContent-length: "
						"%d\r\n\r\n", bytesTotal);
			break;
		default	:
			fprintf(stderr, "ERROR : BUG : Upload::loginToPeer() : invalid type!!!\r\n");
			state	= CON_STATE_CON_ERROR;
			return 0;
	}
	sendStr(buf, strlen(buf));
 	setState(CON_STATE_ACTIVE);
	bufferSend();
	return 1;
}

int Upload::sendBlob()
{
	int		bs;

	if (feof(file) && !(outBuffer->bytesInBuffer()))
	{
		setState(CON_STATE_DONE);
		return 0;
	}

	/*
	 * Add new data from file
	 */

	if (!feof(file))
		outBuffer->bufFRead(file);

	/*
	 * Send
	 */
	bs	= bufferSend();
	
	if (bs != -1)
	{
		bytesLeft	-= bs;
		bytesDone	+= bs;
	}
	else
		setState(CON_STATE_CON_ERROR);

	return bs;
}

void Upload::setCID(char *newcid)
{
	int	i;
	
	for (i = 0; i < 2 * CLIENTID_LENGTH; ++i)
	{
		unsigned char	idx;
		char		hex []	=  {	'0', '1', '2', '3', '4', '5', '6', '7',
					'8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

		if (i % 2)
			idx	= newcid[i / 2] & 0xF;
		else
			idx	= (newcid[i / 2] & 0xF0) / 0x10;

 		aacid[i]	= hex[ idx ];
	}
}

